# coding=utf-8
from mammut.common.lexicon.pos import POSType
from brat_widget.document import *
import copy


class Parents:
    def __init__(self, depth, parents=None):
        self.parents = parents
        self.depth = depth

    def get(self):
        if self.parents is not None:
            return self.parents


class Overlap:
    def __init__(self, overlap_type):
        self.overlap_type = overlap_type


class Entity:
    def __init__(self, entity, parents=None):
        self.entity = entity  # objeto annotation
        self.parents = parents


# permitimos que entities overlap pueda expresar los solapamientos de los entities y tmabien los solapamientos de eventos, simplemente sabemos que los
# entites de eventos solapados son iguales
# convierte el evento solapado e un objeto entitiesoverlap
class EventEntitesOverlap:
    def __init__(self, entity, parents):
        self.entites_overlap = EntitiesOverlap()
        self.entity_original = Entity(entity, parents)
        self.entites_overlap.add(self.entity_original)
        self.entity_copy = Entity(entity, parents)
        self.entites_overlap.add(self.entity_copy)

    def get_entities_overlap(self):
        return self.enties_overlap


class EntitiesOverlap:
    def __init__(self):
        self.entities_list = list()

    def add(self, entity):
        self.entities_list.append(entity)


class EventAnnotationValidator:
    def __init__(self, package, annotation):
        self.package = package
        self.annotation = annotation
        self.valid = True
        self.invalid_type = list()  # event_list y/o entity listy
        self.annotation_lists_by_dimension = dict()
        self.sorted_annotations_list_by_dimension = dict()
        self.valid_sets_by_dimention = dict()
        self.invalid_spans_pairs_by_dimension = dict()
        self.event_annotation_by_dimension = dict()
        self.all_enties_with_overlaped_parents_by_dimension = dict()
        self.overlaped_enties_parents_by_dimension = dict()
        self.has_nested_overlaps = False

        for dimension in self.annotation.documents_by_dimension:
            self.sorted_annotations_list_by_dimension[dimension] = dict()
            self.valid_sets_by_dimention[dimension] = dict()
            self.annotation_lists_by_dimension[dimension] = dict()
            self.annotation_lists_by_dimension[dimension][
                EventsAnnotationsValidator.ENTITIES_LIST
            ] = list()
            self.annotation_lists_by_dimension[dimension][
                EventsAnnotationsValidator.EVENTS_LIST
            ] = list()
            self.event_annotation_by_dimension[dimension] = list()

        self.get_entities_and_events_lists()
        self.sort_and_validate()

    def get_entities_and_events_lists(self):
        for dimension in self.annotation.documents_by_dimension:
            for i in self.annotation.documents_by_dimension[
                dimension
            ].ann_obj.get_textbounds():
                # TODO:no se incluyen las preposiciones, i.type no es un key de descriptor_list
                if i.type in self.package.standard.pos_descriptor_list:
                    if (
                        self.package.standard.pos_descriptor_list[i.type].pos_type
                        != POSType.Event
                    ):
                        self.annotation_lists_by_dimension[dimension][
                            EventsAnnotationsValidator.ENTITIES_LIST
                        ].append(i)
                    else:
                        self.annotation_lists_by_dimension[dimension][
                            EventsAnnotationsValidator.EVENTS_LIST
                        ].append(i)
            for i in self.annotation.documents_by_dimension[
                dimension
            ].ann_obj.get_events():
                self.event_annotation_by_dimension[dimension].append(i)

    # recibe lista de textboundanotations
    # TODO:list_type es indica si es entity o event_list, de ser event list creo que no tiene utilidad
    def is_valid_subset(self, sorted_list, dimension, list_type):
        valid = True
        for i, elem in enumerate(sorted_list[:-1]):
            a = sorted_list[i].spans[0]
            b = sorted_list[i + 1].spans[0]
            seta = set(range(a[0], a[1] + 1))
            setb = set(range(b[0], b[1] + 1))
            # print(valid, seta, setb, seta-setb, setb-seta)
            if dimension not in self.invalid_spans_pairs_by_dimension.keys():
                self.invalid_spans_pairs_by_dimension[dimension] = dict()
            if list_type not in self.invalid_spans_pairs_by_dimension[dimension].keys():
                self.invalid_spans_pairs_by_dimension[dimension][list_type] = list()
            if len(seta & setb) > 1 and len(seta - setb) > 0 and len(setb - seta) > 0:
                invalid_tuple = tuple([sorted_list[i], sorted_list[i + 1]])
                self.invalid_spans_pairs_by_dimension[dimension][list_type].append(
                    invalid_tuple
                )
                valid = False
                # break
        return valid

    def get_nested_overlaped_events(self):
        overlaped_spans = list()
        overlaps_by_dimension = self.all_enties_with_overlaped_parents_by_dimension
        # entity = ovrelap
        for dimension, overlap in overlaps_by_dimension.items():
            if len(overlap):
                overlaped_spans.append(
                    (
                        overlap[0].entities_list[0].entity.text,
                        self.annotation.event["original_text"],
                        dimension,
                    )
                )
        return overlaped_spans

    def get_parents(self, span, dimension):
        # print(span.type, 'de quien obtenemos los padres')
        parents = list()
        # TODO: no deberia incluir este metodo tambien alas entidades solapadas?overlaped_enties_parents_by_dimension
        for event in self.annotation.documents_by_dimension[
            dimension
        ].ann_obj.get_events():
            # TODO:deberia ser en funcion del texto o del id??con el texto se obtienen overlaps raros
            # if span.type in [arg[0] for arg in event.args]:
            if span.id in [arg[1] for arg in event.args]:
                parents.append(
                    self.annotation.documents_by_dimension[
                        dimension
                    ].ann_obj.get_ann_by_id(event.trigger)
                )
        if len(parents):
            for i in parents:
                self.get_parents(i, dimension)
        return parents

    def get_events_arguments(self, dimension):
        # calculamos los padres de todos los spans
        parents = None
        if dimension not in self.all_enties_with_overlaped_parents_by_dimension.keys():
            self.all_enties_with_overlaped_parents_by_dimension[dimension] = list()
        if dimension not in self.overlaped_enties_parents_by_dimension.keys():
            self.overlaped_enties_parents_by_dimension[dimension] = list()
        for entity in self.sorted_annotations_list_by_dimension[dimension][
            EventsAnnotationsValidator.ENTITIES_LIST
        ]:
            parents = self.get_parents(entity, dimension)
            # print('*************')
            if len(parents) >= 2:
                self.has_nested_overlaps = True
                self.all_enties_with_overlaped_parents_by_dimension[dimension].append(
                    EventEntitesOverlap(entity, parents).entites_overlap
                )
                self.valid = False
                # TODO:revisar esto de invalid type, puede ser invalido en varios types
                if EventsAnnotationsValidator.EVENTS_LIST not in self.invalid_type:
                    self.invalid_type.append(EventsAnnotationsValidator.EVENTS_LIST)
                # TODO: si esto pasa marcamos el evento como solpado/invalido??
                # print('&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&')
                # print(parents)
        # si hay overlap de entities en capa1 obtenemos los padres de los solapados
        # y sabemos que los padres son eventos solapados
        if dimension in self.invalid_spans_pairs_by_dimension.keys():
            if (
                EventsAnnotationsValidator.ENTITIES_LIST
                in self.invalid_spans_pairs_by_dimension[dimension].keys()
            ):
                # si entra en este fi es poruqe son entities con solapamiento por eso marcamos como invalidos e indicamos quienes son los parents
                for overlaped_arguments in self.invalid_spans_pairs_by_dimension[
                    dimension
                ][EventsAnnotationsValidator.ENTITIES_LIST]:
                    parents = None
                    parents_list = list()
                    enties_overlap = EntitiesOverlap()
                    for argument in overlaped_arguments:
                        print("##### when entie overlap")
                        parents = self.get_parents(argument, dimension)
                        parents_list = parents_list + parents
                        enties_overlap.add(Entity(argument, parents))
                    # TODO: duplicare algo de la data de all_enties_with_overlaped_parents_by_dimension?
                    if len(parents_list) >= 2:
                        self.overlaped_enties_parents_by_dimension[dimension].append(
                            enties_overlap
                        )
                        self.has_nested_overlaps = True
                        self.valid = False
                        if (
                            EventsAnnotationsValidator.EVENTS_LIST
                            not in self.invalid_type
                        ):
                            self.invalid_type.append(
                                EventsAnnotationsValidator.EVENTS_LIST
                            )
                        print("overlaped events")

    def sort_and_validate(self):
        for dimension in self.sorted_annotations_list_by_dimension:
            # print(dimension)
            self.sorted_annotations_list_by_dimension[dimension][
                EventsAnnotationsValidator.ENTITIES_LIST
            ] = self.bubble_sort(
                self.annotation_lists_by_dimension[dimension][
                    EventsAnnotationsValidator.ENTITIES_LIST
                ]
            )
            self.sorted_annotations_list_by_dimension[dimension][
                EventsAnnotationsValidator.EVENTS_LIST
            ] = self.bubble_sort(
                self.annotation_lists_by_dimension[dimension][
                    EventsAnnotationsValidator.EVENTS_LIST
                ]
            )
            self.valid_sets_by_dimention[dimension][
                EventsAnnotationsValidator.ENTITIES_LIST
            ] = self.is_valid_subset(
                self.sorted_annotations_list_by_dimension[dimension][
                    EventsAnnotationsValidator.ENTITIES_LIST
                ],
                dimension,
                EventsAnnotationsValidator.ENTITIES_LIST,
            )
            # TODO: esta validacion de entities de eventos solapados no aporta informacion valiosa???
            self.valid_sets_by_dimention[dimension][
                EventsAnnotationsValidator.EVENTS_LIST
            ] = self.is_valid_subset(
                self.sorted_annotations_list_by_dimension[dimension][
                    EventsAnnotationsValidator.EVENTS_LIST
                ],
                dimension,
                EventsAnnotationsValidator.EVENTS_LIST,
            )
            # NOTE: despues de validar si es invalido como se hacian antes, ahora validamos contran los argumentos de los eventos
            self.get_events_arguments(dimension)  # TODO:esto puede ir e otro metodo??
        for dimension in self.valid_sets_by_dimention:
            for key, val in self.valid_sets_by_dimention[dimension].items():
                if not val:  # un valor falso, que e slo mismo que un set invalido
                    self.valid = False
                    # print(key)
                    if key not in self.invalid_type:
                        self.invalid_type.append(key)
                        # break

    def contains(self, tupA, tupB):
        # print('a en b', tupA, tupB, tupA[0] >= tupB[0] and tupA[1]<=tupB[1])
        return tupA[0] >= tupB[0] and tupA[1] <= tupB[1]

    def bubble_sort(self, array):
        # print('#########################\nunsorted\n', [i.spans[0] for i in array])
        index = len(array) - 1
        while index >= 0:
            for j in range(index):
                # A = array[j]
                # B = array[j+1]
                # print('pair', array[j].spans[0], array[j+1].spans[0])
                if (
                    array[j].spans[0][0] > array[j + 1].spans[0][0]
                ):  # comparacion de primer numero de la tupla
                    if not self.contains(array[j].spans[0], array[j + 1].spans[0]):
                        array[j], array[j + 1] = array[j + 1], array[j]
                elif array[j].spans[0][0] < array[j + 1].spans[0][0]:
                    if self.contains(array[j + 1].spans[0], array[j].spans[0]):
                        array[j], array[j + 1] = array[j + 1], array[j]
                elif array[j].spans[0][0] == array[j + 1].spans[0][0]:
                    if self.contains(
                        array[j + 1].spans[0], array[j].spans[0]
                    ):  # segundo elemento contenido enel primero?
                        # if not contains(array[j], array[j+1]):#primer elemento contenido enel segundo?
                        array[j], array[j + 1] = array[j + 1], array[j]
                        # print([i.spans[0] for i in array])

            index -= 1
        # print('\nsorted\n', [i.spans[0] for i in array])
        return array

    # esto computa casi lo mismo que is_valid_subset
    def get_overlap_spans(self, sorted_list):
        overlap_pairs = list()
        valid = True
        for i, elem in enumerate(sorted_list[:-1]):
            a = sorted_list[i].spans[0]
            b = sorted_list[i + 1].spans[0]
            seta = set(range(a[0], a[1] + 1))
            setb = set(range(b[0], b[1] + 1))
            # print(valid, seta, setb, seta-setb, setb-seta)
            # print(valid, seta, setb, seta-setb, setb-seta)
            if seta & setb and len(seta - setb) > 0 and len(seta - setb) > 0:
                # overlap_pairs.append(sorted_list[i].type +'-'+ sorted_list[i+1].type)
                overlap_pairs.append((sorted_list[i], sorted_list[i + 1]))
        return overlap_pairs


class EventsAnnotationsValidator:
    ENTITIES_LIST = "entities_list"
    EVENTS_LIST = "events_list"

    def __init__(self, corpus, package):
        self.package = package
        self.event_annotation_validator_list = list()
        self.invalid_events = list()
        self.valid_events = list()
        self.invalid_by_dimension = dict()
        self.invalid_annotations_pairs_count = dict()
        self.overlap_spans_list = list()

        for event_ann in self.package.annotations.annotation_event_by_corpus[
            corpus.sheet_title
        ]:
            eavalidator = EventAnnotationValidator(self.package, event_ann)
            self.event_annotation_validator_list.append(eavalidator)
            if eavalidator.valid:
                self.valid_events.append(eavalidator)
            else:
                self.invalid_events.append(eavalidator)
            for dimension in eavalidator.valid_sets_by_dimention:
                # invalidos por dimension
                if dimension not in self.invalid_by_dimension.keys():
                    self.invalid_by_dimension[dimension] = {}
                if (
                    EventsAnnotationsValidator.EVENTS_LIST
                    not in self.invalid_by_dimension[dimension].keys()
                ):
                    self.invalid_by_dimension[dimension][
                        EventsAnnotationsValidator.EVENTS_LIST
                    ] = list()
                if (
                    EventsAnnotationsValidator.ENTITIES_LIST
                    not in self.invalid_by_dimension[dimension].keys()
                ):
                    self.invalid_by_dimension[dimension][
                        EventsAnnotationsValidator.ENTITIES_LIST
                    ] = list()
                # se toman en cuenta eventos para analisis
                if eavalidator.has_nested_overlaps:
                    # solo aqi porque son de tipo event_list
                    self.invalid_by_dimension[dimension][
                        EventsAnnotationsValidator.EVENTS_LIST
                    ] = (
                        self.invalid_by_dimension[dimension][
                            EventsAnnotationsValidator.EVENTS_LIST
                        ]
                        + eavalidator.all_enties_with_overlaped_parents_by_dimension[
                            dimension
                        ]
                    )
                # se toman en cuenta los entities para analisis
                if not eavalidator.valid_sets_by_dimention[dimension][
                    EventsAnnotationsValidator.ENTITIES_LIST
                ]:
                    overlap_pairs = eavalidator.get_overlap_spans(
                        eavalidator.sorted_annotations_list_by_dimension[dimension][
                            EventsAnnotationsValidator.ENTITIES_LIST
                        ]
                    )
                    enties_overlap = EntitiesOverlap()
                    for pair in overlap_pairs:
                        for entity_annotation in pair:
                            enties_overlap.add(Entity(entity_annotation))
                    self.invalid_by_dimension[dimension][
                        EventsAnnotationsValidator.ENTITIES_LIST
                    ].append(enties_overlap)
                    overlap_span_values = (
                        overlap_pairs[0][1].spans[0][0],
                        overlap_pairs[0][0].spans[0][1],
                    )
                    self.overlap_spans_list.append(
                        (
                            eavalidator.annotation.event["original_text"][
                                overlap_span_values[0] : overlap_span_values[1]
                            ],
                            eavalidator.annotation.event["original_text"],
                            dimension,
                        )
                    )
            if eavalidator.has_nested_overlaps:
                self.overlap_spans_list = (
                    self.overlap_spans_list + eavalidator.get_nested_overlaped_events()
                )

        for (
            key,
            dimension_values,
        ) in self.invalid_by_dimension.items():  # echange, {event_list:[],entitlist:[]}
            for list_type, list_element in dimension_values.items():
                for entitie_overlap in list_element:
                    ann_pair_text = (
                        entitie_overlap.entities_list[0].entity.type
                        + "-"
                        + entitie_overlap.entities_list[1].entity.type
                    )
                    if ann_pair_text not in self.invalid_annotations_pairs_count:
                        self.invalid_annotations_pairs_count[ann_pair_text] = dict()
                        self.invalid_annotations_pairs_count[ann_pair_text] = 1
                    else:
                        self.invalid_annotations_pairs_count[ann_pair_text] += 1

    def get_annotation_params_copy(self, annotation, dimension):
        params = dict()
        params["text_bounds"] = list()
        for i in annotation.documents_by_dimension[dimension].ann_obj.get_textbounds():
            params["text_bounds"].append(i)
        params["events"] = list()
        for i in annotation.documents_by_dimension[dimension].ann_obj.get_events():
            params["events"].append(i)
        params["attributes"] = list()
        for i in annotation.documents_by_dimension[dimension].ann_obj.get_attributes():
            params["attributes"].append(i)
        return params

    def restore_from_params(self, annotation, dimension, params):
        annotation.initialize_configuration(dimension)
        annotation.initialize_document(dimension, annotation.event)
        document = annotation.documents_by_dimension[dimension]
        for text_bound in params["text_bounds"]:
            document.ann_obj.add_annotation(text_bound)
        for event in params["events"]:
            document.ann_obj.add_annotation(event)
        for attributes in params["attributes"]:
            document.ann_obj.add_annotation(attributes)

    def visualize_overlap_from_event_validator(
        self, ann_obj_validator, selected_dimension
    ):
        visualizers_list = list()
        for dimension, dim_vals in ann_obj_validator.valid_sets_by_dimention.items():
            # print(dimension, dim_vals, '<<<<<<<<<<<<<<<<<<<<<<<>>>>>>')
            for type_list, elem_type in dim_vals.items():
                if False == elem_type and dimension == selected_dimension:
                    params_copy = self.get_annotation_params_copy(
                        ann_obj_validator.annotation, dimension
                    )
                    # print('dibujamos data de ', type_list, dimension)
                    ann_obj_validator.annotation.initialize_configuration(dimension)
                    ann_obj_validator.annotation.initialize_document(
                        dimension, ann_obj_validator.annotation.event
                    )
                    document = ann_obj_validator.annotation.documents_by_dimension[
                        dimension
                    ]
                    overlap_pairs = ann_obj_validator.get_overlap_spans(
                        ann_obj_validator.sorted_annotations_list_by_dimension[
                            dimension
                        ][type_list]
                    )
                    # print(overlap_pairs)
                    for i in overlap_pairs[0]:
                        # print(i)
                        annotation = TextBoundAnnotationWithText(
                            i.spans, i.id, i.type, i.text
                        )
                        # print('overlap_pair', annotation)
                        document.ann_obj.add_annotation(annotation)
                    visualizers_list.append(
                        ann_obj_validator.annotation.get_visualizer(dimension, None)
                    )
                    self.restore_from_params(
                        ann_obj_validator.annotation, dimension, params_copy
                    )
        return visualizers_list

    def get_event_appearance(self, ann_obj_validator, elem_type):
        count = 0
        annotation_events = [
            i.annotation for i in self.invalid_events + self.valid_events
        ]
        for index, event_ann in enumerate(annotation_events):
            for dimension in event_ann.documents_by_dimension:
                for i in event_ann.documents_by_dimension[
                    dimension
                ].ann_obj.get_textbounds():
                    if i.type == elem_type:
                        count += 1
        return count

    def get_event_overlap_and_overall_count(self, ann_obj_validator):
        overlap_pair_type_count = dict()
        overlap_pair_type = None
        for (
            dimension
        ) in (
            self.invalid_by_dimension
        ):  # cualquier dict que tenga los key de dimensiones
            for invalid_type in ann_obj_validator.invalid_type:
                overlap_pairs = ann_obj_validator.get_overlap_spans(
                    ann_obj_validator.sorted_annotations_list_by_dimension[dimension][
                        invalid_type
                    ]
                )
                if len(overlap_pairs):
                    overlap_pair_type = (
                        overlap_pairs[0][0].type + "-" + overlap_pairs[0][1].type
                    )
                    for overlap_pair in overlap_pairs:
                        appearance_count = dict()
                        for i in overlap_pair:
                            appearance_count[i.type] = self.get_event_appearance(
                                ann_obj_validator, i.type
                            )
                            overlap_pair_type_count[
                                overlap_pair_type
                            ] = appearance_count
        return overlap_pair_type_count


class CorpusAnnotationsValidator:
    def __init__(self, package):
        self.annotations_validator_by_corpus = dict()
        for corpus in package.corpus_map.get_all_corpus():
            if corpus.annotable:
                annotations_validator = EventsAnnotationsValidator(corpus, package)
                self.annotations_validator_by_corpus[
                    corpus.sheet_title
                ] = annotations_validator



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: apps/WorkingMemory.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='apps/WorkingMemory.proto',
  package='com.mammut.interpreter.memory',
  syntax='proto3',
  serialized_options=_b('B\023WorkingMemoryProtos'),
  serialized_pb=_b('\n\x18\x61pps/WorkingMemory.proto\x12\x1d\x63om.mammut.interpreter.memory\"\xc2\x05\n\x12WorkingMemoryValue\x12\x45\n\x04type\x18\x01 \x01(\x0e\x32\x37.com.mammut.interpreter.memory.WorkingMemoryValue.types\x12U\n\x0c\x63\x61ridanility\x18\x02 \x01(\x0e\x32?.com.mammut.interpreter.memory.WorkingMemoryValue.cardinalities\x12\x12\n\x08valueInt\x18\x03 \x01(\x05H\x00\x12\x13\n\tvalueLong\x18\x04 \x01(\x03H\x00\x12\x15\n\x0bvalueString\x18\x05 \x01(\tH\x00\x12\x13\n\tvalueBool\x18\x06 \x01(\x08H\x00\x12\x15\n\x0bvalueDouble\x18\x07 \x01(\x01H\x00\x12\x14\n\nvalueOther\x18\x08 \x01(\tH\x00\x12\x0f\n\x07valueRS\x18\t \x03(\t\x12\x0f\n\x07valueRI\x18\n \x03(\x03\x12\x0f\n\x07valueRL\x18\x0b \x03(\x03\x12\x0f\n\x07valueRB\x18\x0c \x03(\x08\x12\x0f\n\x07valueRD\x18\r \x03(\x01\x12\x0f\n\x07valueRO\x18\x0e \x03(\t\"\xeb\x01\n\x05types\x12\n\n\x06STRING\x10\x00\x12\x07\n\x03INT\x10\x01\x12\x08\n\x04LONG\x10\x02\x12\x08\n\x04\x42OOL\x10\x03\x12\n\n\x06\x44OUBLE\x10\x04\x12\x11\n\rZONEDDATETIME\x10\x05\x12\r\n\tLOCALDATE\x10\x06\x12\r\n\tLOCALTIME\x10\x07\x12\x0e\n\nATTACHMENT\x10\x10\x12\x0f\n\x0bGATTACHMENT\x10\x11\x12\x0f\n\x0bGEOLOCATION\x10\x12\x12\x10\n\x0cOPENINGHOURS\x10\x13\x12\x11\n\rSTRINGTOLOWER\x10\x14\x12\x11\n\rSTRINGTOUPPER\x10\x15\x12\x12\n\x0eTYPEDESCRIPTOR\x10\x16\".\n\rcardinalities\x12\n\n\x06SINGLE\x10\x00\x12\x08\n\x04LIST\x10\x01\x12\x07\n\x03SET\x10\x02\x42\x08\n\x06valuesB\x15\x42\x13WorkingMemoryProtosb\x06proto3')
)



_WORKINGMEMORYVALUE_TYPES = _descriptor.EnumDescriptor(
  name='types',
  full_name='com.mammut.interpreter.memory.WorkingMemoryValue.types',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STRING', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LONG', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BOOL', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DOUBLE', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ZONEDDATETIME', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOCALDATE', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOCALTIME', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ATTACHMENT', index=8, number=16,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GATTACHMENT', index=9, number=17,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GEOLOCATION', index=10, number=18,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OPENINGHOURS', index=11, number=19,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STRINGTOLOWER', index=12, number=20,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='STRINGTOUPPER', index=13, number=21,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPEDESCRIPTOR', index=14, number=22,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=473,
  serialized_end=708,
)
_sym_db.RegisterEnumDescriptor(_WORKINGMEMORYVALUE_TYPES)

_WORKINGMEMORYVALUE_CARDINALITIES = _descriptor.EnumDescriptor(
  name='cardinalities',
  full_name='com.mammut.interpreter.memory.WorkingMemoryValue.cardinalities',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SINGLE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LIST', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SET', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=710,
  serialized_end=756,
)
_sym_db.RegisterEnumDescriptor(_WORKINGMEMORYVALUE_CARDINALITIES)


_WORKINGMEMORYVALUE = _descriptor.Descriptor(
  name='WorkingMemoryValue',
  full_name='com.mammut.interpreter.memory.WorkingMemoryValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='caridanility', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.caridanility', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueInt', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueInt', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueLong', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueLong', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueString', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueString', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueBool', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueBool', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueDouble', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueDouble', index=6,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueOther', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueOther', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueRS', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueRS', index=8,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueRI', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueRI', index=9,
      number=10, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueRL', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueRL', index=10,
      number=11, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueRB', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueRB', index=11,
      number=12, type=8, cpp_type=7, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueRD', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueRD', index=12,
      number=13, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='valueRO', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.valueRO', index=13,
      number=14, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _WORKINGMEMORYVALUE_TYPES,
    _WORKINGMEMORYVALUE_CARDINALITIES,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='values', full_name='com.mammut.interpreter.memory.WorkingMemoryValue.values',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=60,
  serialized_end=766,
)

_WORKINGMEMORYVALUE.fields_by_name['type'].enum_type = _WORKINGMEMORYVALUE_TYPES
_WORKINGMEMORYVALUE.fields_by_name['caridanility'].enum_type = _WORKINGMEMORYVALUE_CARDINALITIES
_WORKINGMEMORYVALUE_TYPES.containing_type = _WORKINGMEMORYVALUE
_WORKINGMEMORYVALUE_CARDINALITIES.containing_type = _WORKINGMEMORYVALUE
_WORKINGMEMORYVALUE.oneofs_by_name['values'].fields.append(
  _WORKINGMEMORYVALUE.fields_by_name['valueInt'])
_WORKINGMEMORYVALUE.fields_by_name['valueInt'].containing_oneof = _WORKINGMEMORYVALUE.oneofs_by_name['values']
_WORKINGMEMORYVALUE.oneofs_by_name['values'].fields.append(
  _WORKINGMEMORYVALUE.fields_by_name['valueLong'])
_WORKINGMEMORYVALUE.fields_by_name['valueLong'].containing_oneof = _WORKINGMEMORYVALUE.oneofs_by_name['values']
_WORKINGMEMORYVALUE.oneofs_by_name['values'].fields.append(
  _WORKINGMEMORYVALUE.fields_by_name['valueString'])
_WORKINGMEMORYVALUE.fields_by_name['valueString'].containing_oneof = _WORKINGMEMORYVALUE.oneofs_by_name['values']
_WORKINGMEMORYVALUE.oneofs_by_name['values'].fields.append(
  _WORKINGMEMORYVALUE.fields_by_name['valueBool'])
_WORKINGMEMORYVALUE.fields_by_name['valueBool'].containing_oneof = _WORKINGMEMORYVALUE.oneofs_by_name['values']
_WORKINGMEMORYVALUE.oneofs_by_name['values'].fields.append(
  _WORKINGMEMORYVALUE.fields_by_name['valueDouble'])
_WORKINGMEMORYVALUE.fields_by_name['valueDouble'].containing_oneof = _WORKINGMEMORYVALUE.oneofs_by_name['values']
_WORKINGMEMORYVALUE.oneofs_by_name['values'].fields.append(
  _WORKINGMEMORYVALUE.fields_by_name['valueOther'])
_WORKINGMEMORYVALUE.fields_by_name['valueOther'].containing_oneof = _WORKINGMEMORYVALUE.oneofs_by_name['values']
DESCRIPTOR.message_types_by_name['WorkingMemoryValue'] = _WORKINGMEMORYVALUE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

WorkingMemoryValue = _reflection.GeneratedProtocolMessageType('WorkingMemoryValue', (_message.Message,), dict(
  DESCRIPTOR = _WORKINGMEMORYVALUE,
  __module__ = 'apps.WorkingMemory_pb2'
  # @@protoc_insertion_point(class_scope:com.mammut.interpreter.memory.WorkingMemoryValue)
  ))
_sym_db.RegisterMessage(WorkingMemoryValue)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# coding=utf-8
import os
import configparser
import datetime
import logging.config
from pkg_resources import resource_string
from pytz import reference

__version__ = "0.1.0"

__all__ = ["apps", "data", "common", "linguistics", "models", "visualization"]

ZONED_DATETIME_FORMATTER = "%Y-%m-%dT%H:%M:%S%z"
LOCAL_DATE_FORMATTER = "%Y-%m-%d"
LOCAL_TIME_FORMATTER = "%H:%M:%S"
TIME_SEPARATOR = ":"
TIME_ALTERNATIVE_SEPARATOR = ";"

LOCAL_TIMEZONE = reference.LocalTimezone()


def get_timezone_aware_datetime_min():
    return datetime.datetime.fromtimestamp(0, LOCAL_TIMEZONE)


def get_timezone_aware_datetime_now():
    return datetime.datetime.now(LOCAL_TIMEZONE)


def str_format_datetime(dt):
    if not dt.tzinfo:
        dt = dt.replace(tzinfo=LOCAL_TIMEZONE)
    ncoldts = dt.strftime(ZONED_DATETIME_FORMATTER)
    wcoldts = ncoldts[:-2] + ":" + ncoldts[-2:]
    return wcoldts


def str_parse_datetime(strdt):
    if len(strdt) <= 19:
        strdt = strdt + get_str_format_timezone_aware_datetime_now()[19:]
    elif len(strdt) <= 20:
        strdt = strdt + get_str_format_timezone_aware_datetime_now()[20:]
    strdt = strdt[:-3] + strdt[-2:]
    dt = datetime.datetime.strptime(strdt, ZONED_DATETIME_FORMATTER)
    return dt


def get_str_format_timezone_aware_datetime_now():
    return str_format_datetime(get_timezone_aware_datetime_now())


ES_CONFIGURATIONS = configparser.ConfigParser()
ES_CONFIGURATIONS.read_string(
    resource_string("mammut.resources", "elastic.ini").decode("utf-8")
)
ES_CONFIGURATIONS.read(["data/elastic.ini", "elastic.ini"])

IMAGES_CONFIGURATION = configparser.ConfigParser()
IMAGES_CONFIGURATION.read_string(
    resource_string("mammut.resources", "models.ini").decode("utf-8")
)
IMAGES_CONFIGURATION.read(["data/models.ini", "models.ini"])

if os.environ.get("ELASTIC_HOST") is not None:
    ES_HOST = os.environ.get("ELASTIC_HOST")
else:
    ES_HOST = ES_CONFIGURATIONS["DEFAULT"]["host"]
if os.environ.get("ELASTIC_PORT") is not None:
    ES_PORT = int(os.environ.get("ELASTIC_PORT"))
else:
    ES_PORT = int(ES_CONFIGURATIONS["DEFAULT"]["port"])
if os.environ.get("ELASTIC_USE_SSL") is not None:
    ES_USE_SSL = bool(os.environ.get("ELASTIC_USE_SSL"))
else:
    ES_USE_SSL = bool(ES_CONFIGURATIONS["DEFAULT"]["use_ssl"])
if os.environ.get("ELASTIC_HTTP_AUTH") is not None:
    ES_HTTP_AUTH = os.environ.get("ELASTIC_HTTP_AUTH")
else:
    ES_HTTP_AUTH = ES_CONFIGURATIONS["DEFAULT"]["http_auth"]

ES_HOSTS = [
    {"host": ES_HOST, "port": ES_PORT, "use_ssl": ES_USE_SSL, "http_auth": ES_HTTP_AUTH}
]
CONCORDANCE_QUERY_SIZE = int(ES_CONFIGURATIONS["DEFAULT"]["concordance_query_size"])

BASE_MODEL_VISON_DIR = IMAGES_CONFIGURATION["IMAGES"]["base_dir"]

LOGGING_CONFIGURATION = configparser.ConfigParser()
LOGGING_CONFIGURATION.read_string(
    resource_string("mammut.resources", "logging.ini").decode("utf-8")
)
logging.config.fileConfig(LOGGING_CONFIGURATION)
if os.path.exists("data/logging.ini"):
    logging.config.fileConfig("data/logging.ini")
if os.path.exists("logging.ini"):
    logging.config.fileConfig("logging.ini")



from mammut.common.lexicon.functional import FunctionalLemmaDescriptor


class NonFunctionalLemmaDescriptor:
    LEMMA_COLUMN_NAME = "lemma"
    HIDE_COLUMN_NAME = "hide"
    REGIONAL_SETTINGS_LEMMA_COLUMN_NAME = "regional-settings-lemma"
    POS_PARADIGM_COLUMN_NAME = "pos-paradigm"
    DEFAULT_START_CELL = "A1"
    DEFAULT_END_CELL = "AF"

    def __init__(self, dataframe_row, standard):
        self.lemma = dataframe_row[FunctionalLemmaDescriptor.LEMMA_COLUMN_NAME].lower()
        self.regional_settings_lemma = dataframe_row[
            FunctionalLemmaDescriptor.REGIONAL_SETTINGS_LEMMA_COLUMN_NAME
        ]
        self.pos_paradigm = dataframe_row[
            FunctionalLemmaDescriptor.POS_PARADIGM_COLUMN_NAME
        ]



# coding=utf-8

from ipywidgets import *
from django.template.loader import get_template

disambiguation_template = get_template("disambiguation.html")
tokens_type_resume_template = get_template("token_type_resume.html")


def disambiguation_widget_section(package):
    def definition_selector(defintion_set_entry_id, morpheme_info):
        definitions_options = {}
        selected_definition_id = {"value": None}
        definitions_set_entries = package.dictionary.get_definitions_entries_as_list(
            defintion_set_entry_id
        )

        def get_disambiguation_aplication_html():
            definitions_dict = {}
            current_definition = None
            for index, definition_entry in enumerate(definitions_set_entries):
                if package.dictionary.is_disambiguation_selected(
                    morpheme_info,
                    package.dictionary.get_definition_complete_id(
                        defintion_set_entry_id, index
                    ),
                ):
                    definitions_dict[
                        "def " + str(index + 1) + " - actual"
                    ] = definition_entry  # TODO mejorar estos string, constantes?
                    current_definition = definition_entry
                else:
                    definitions_dict["def " + str(index + 1)] = definition_entry
            disambiguation_application_html = disambiguation_template.render(
                {
                    "definition_sets": definitions_dict,
                    "current_definition": current_definition,
                }
            )
            disambiguation_application = widgets.HTML(
                value=disambiguation_application_html
            )
            return disambiguation_application

        for index, i in enumerate(definitions_set_entries):
            definitions_options[
                "def " + str(index + 1)
            ] = package.dictionary.get_definition_complete_id(
                defintion_set_entry_id, index
            )
        disambiguation_application = get_disambiguation_aplication_html()
        defintions_radio_buttons = widgets.RadioButtons(
            options=definitions_options,
            value=None,
            description="select one definition:",
            layout=Layout(width="900px"),
        )

        def defintions_radio_buttons_change(change):
            if change["new"] is not None:
                selected_definition_id["value"] = change["new"]

        defintions_radio_buttons.observe(defintions_radio_buttons_change, names="value")
        save_definition_button = widgets.Button(
            description="Guardar Definition",
            button_style="",
            tooltip="Guardar Defiinition",
            icon="",
            layout=Layout(width="200px"),
        )

        def handle_save_definition(sender):
            if selected_definition_id["value"] is not None:
                package.dictionary.save_lema_disambiguation(
                    morpheme_info, selected_definition_id["value"]
                )
                tokens_selector_elements.children = [
                    get_disambiguation_aplication_html(),
                    defintions_radio_buttons,
                    save_definition_button,
                ]

        save_definition_button.on_click(handle_save_definition)

        tokens_selector_elements = widgets.VBox(
            [
                disambiguation_application,
                defintions_radio_buttons,
                save_definition_button,
            ]
        )

        return tokens_selector_elements

    def token_selector(event_annotation):
        (
            tokens_dict,
            tokens_info_dict,
            tokens_type_resume,
        ) = package.dictionary.get_event_tokens_dict_and_info(event_annotation)

        tokens_type_resume_html = tokens_type_resume_template.render(
            {"token_types": tokens_type_resume}
        )
        tokens_type_resume = widgets.HTML(value=tokens_type_resume_html)
        tokens_dict["None"] = "None"
        tokens = [i for i in tokens_dict]
        tokens_dropdown = widgets.Dropdown(
            options=tokens,
            value="None",
            description="select one token:",
            layout=Layout(width="500px"),
        )

        def tokens_dropdown_change(change):
            if change["new"] is not "None":
                tokens_selector_elements.children = [
                    tokens_dropdown,
                    definition_selector(
                        tokens_dict[change["new"]], tokens_info_dict[change["new"]]
                    ),
                ]
            else:
                tokens_selector_elements.children = [tokens_dropdown]

        tokens_dropdown.observe(tokens_dropdown_change, names="value")

        tokens_selector_elements = widgets.VBox([tokens_type_resume, tokens_dropdown])

        return tokens_selector_elements

    def event_selector(corpus):
        annotation_object = package.annotations
        all_events = annotation_object.get_event_messages(corpus)
        annotation_events = annotation_object.get_annotation_event_messages(corpus)
        all_events_dict = {}
        for annotaion_event_message in annotation_events:
            all_events_dict[
                annotaion_event_message.event["original_text"]
            ] = annotaion_event_message
        all_events_dict["None"] = "None"

        events_dropdown = widgets.Dropdown(
            options=all_events_dict,
            value="None",
            description="all events:",
            layout=Layout(width="800px"),
        )

        def events_dropdown_change(change):
            if change["new"] is not "None":
                disambiguation_widget.children = [
                    corpus_dropdown,
                    events_dropdown,
                    token_selector(change["new"]),
                ]
            else:
                disambiguation_widget.children = [corpus_dropdown, events_dropdown]

        events_dropdown.observe(events_dropdown_change, names="value")
        return events_dropdown

    all_corpus = package.corpus_map.get_all_corpus()
    all_corpus_dict = {}
    for i in all_corpus:
        all_corpus_dict[i.sheet_title] = i
    all_corpus_dict["None"] = "None"
    corpus_dropdown = widgets.Dropdown(
        options=all_corpus_dict,
        value="None",
        description="select corpus:",
        layout=Layout(width="500px"),
    )

    def corpus_dropdown_change(change):
        if change["new"] is not "None":
            disambiguation_widget.children = [
                corpus_dropdown,
                event_selector(change["new"]),
            ]
        else:
            disambiguation_widget.children = [corpus_dropdown]

    corpus_dropdown.observe(corpus_dropdown_change, names="value")

    disambiguation_widget = widgets.VBox([corpus_dropdown])
    return disambiguation_widget



import string
import itertools
import pandas as pd
from mammut.assessment.problem import Problem
from mammut.curriculum.models.sentencepiece import SentencepieceBaseModel
from typing import List, Tuple, Dict
import re


class SentencepieceProblemMFT(Problem):
    """Concrete class to evaluate Google sentencepiece
    trained models.

    Test a minimal functionality expected in the
    discovered vocabulary.

    Note: receives string values from storage manager
        retrieved dataset.
    """

    def __init__(
        self,
        course_id: str,
        assessment_id: str,
        schedule: str,
        problem_id: str,
        problem_type: str,
        parameters: str,
        language_competence: str,
        weight: str,
        retry: str,
        observations: str,
    ):
        """
        Args:
            course_id(str): id of the course to which this problem belongs.
            assessment_id(str): id of the assessment that bundles this problem.
            schedule(str): lesson schedule to execute this problem.
            problem_id(str): id for this problem.
            problem_type(str): problem type enumeration string.
            parameters (str): JSON string parameters for this problem.
            language_competence(str): language competence enumeration value for this problem.
            weight(str): float string value for the weight of this particular problem
                in the overall course score.
            observations(str): observations string. Only descriptive.
        """
        Problem.__init__(
            self,
            course_id,
            assessment_id,
            schedule,
            problem_id,
            problem_type,
            parameters,
            language_competence,
            weight,
            retry,
            observations,
        )
        """
            Template base tags specified for this problem.
            For example:
                {root} {flex1} {flex2}
        """
        self._template_base: str = ""
        """
            _filled_template attribute contains the lists of values for the 
            template tags.
        """
        self._filled_template: List[List[str]] = []
        """
            _rendered_segmented_words contains a list of the expected segmentation
            strings given the template arguments for this test. 
            
            For example:
                [('est', 'uvi', 'era'),
                 ('est', 'uvi', 'ese'),
                 ('est', 'uvi', 'eras'),
                 ('est', 'uvi', 'eses'),
                 ('est', 'uvi', 'era'),
                 ('est', 'uvi', 'ese'),
                 ('est', 'uvi', 'eramos'),
                 ('est', 'uvi', 'esemos'),
                 ('est', 'uvi', 'eran'),
                 ('est', 'uvi', 'esen')]
        """
        self._rendered_segmented_words: List[Tuple[str]] = []
        """
            _rendered_words contains the list of rendered words for the template
        """
        self._rendered_words: List[str] = []
        self._parse_parameters()
        self._render_segmented_words()
        self._render_words()

    def _parse_parameters(self):
        template_base_rgx = re.compile("\{(.*?)\}+")
        self._template_base = self.parameters["template_base"]
        """
            Take the words from the string tags.
            Joins them by comma, converts to a list by splitting.
            {root} {flex} -> ['root', 'flex']
        """
        template_base_words_list = ",".join(
            template_base_rgx.findall(self.parameters["template_base"])
        ).split(",")
        for word in template_base_words_list:
            self._filled_template.append(self.parameters[word])

    def _render_segmented_words(self):
        """This method computes by segments the rendered words
                from the 'render_words' method.

            Example:
                input: [['invent', 'rebaj', 'plasm', 'labr', 'lav',
                    'habit', 'habl'], ['o', 'as', 'a', 'amos', 'áis', 'an']]
                output: [('invent', 'o'), ('invent', 'as'), ('invent', 'a'),
                    ('invent', 'amos'), ('invent', 'áis'), ('invent', 'an'),
                    ('rebaj', 'o'), ('rebaj', 'as'), ('rebaj', 'a'), ('rebaj', 'amos'),
                    ('rebaj', 'áis'), ('rebaj', 'an'), ('plasm', 'o'), ('plasm', 'as'),
                    ('plasm', 'a'), ('plasm', 'amos'), ('plasm', 'áis'), ('plasm', 'an'),
                    ('labr', 'o'), ('labr', 'as'), ('labr', 'a'), ('labr', 'amos'),
                    ('labr', 'áis'), ('labr', 'an'), ('lav', 'o'), ('lav', 'as'),
                    ('lav', 'a'), ('lav', 'amos'), ('lav', 'áis'), ('lav', 'an'),
                    ('habit', 'o'), ('habit', 'as'), ('habit', 'a'), ('habit', 'amos'),
                    ('habit', 'áis'), ('habit', 'an'), ('habl', 'o'), ('habl', 'as'),
                    ('habl', 'a'), ('habl', 'amos'), ('habl', 'áis'), ('habl', 'an')]

            Return:
                Returns a list of segmented words.
        """
        words = [m for m in itertools.product(*self._filled_template)]
        render_seg_words = [word for word in words]
        self._rendered_segmented_words = render_seg_words

    def _render_words(self):
        """This method a list of the words that can be built with the template.

           Example:
               input: [['invent', 'rebaj', 'plasm', 'labr', 'lav',
                   'habit', 'habl'], ['o', 'as', 'a', 'amos', 'áis', 'an']]
               output: ['invento', 'inventas', 'inventa', 'inventamos',
                   'inventáis', 'inventan', 'rebajo', 'rebajas', 'rebaja',
                   'rebajamos', 'rebajáis', 'rebajan', 'plasmo', 'plasmas',
                   'plasma', 'plasmamos', 'plasmáis', 'plasman', 'labro',
                   'labras', 'labra', 'labramos', 'labráis', 'labran', 'lavo',
                   'lavas', 'lava', 'lavamos', 'laváis', 'lavan', 'habito',
                   'habitas', 'habita', 'habitamos', 'habitáis', 'habitan',
                   'hablo', 'hablas', 'habla', 'hablamos', 'habláis', 'hablan']

           Return:
               Returns a list of words built.
       """
        words = [m for m in itertools.product(*self._filled_template)]
        self._rendered_words = ["".join(word) for word in words]

    def _test_run(self, subword_tokenizer, whitespace: bool = False):
        """Runs a test to check that the segmentation built into
            a template is equal to the segmentation of the rendered
            word made by a subword_tokenizer.

            This method works with the 'template method' which defines
            separate morphemes of words, and the 'render_words method'
            which builds the words.

            The test consists of taking as an argument a morpheme tokenizer,
            that takes the separate morphemes of a word. Then the tokenizer
            is used to segment the rendered word and the result is compared
            with the word segmented by the template method.

            Args:
                subword_tokenizer: an object that represents a tokenization model.
                whitespace(bool): False by default. If True, at the
                    time of testing it ignores the white space produced
                    by SentencePiece's subword_tokenizer.
        """
        test_successful = list()
        test_failed = list()
        for num_test in range(len(self._rendered_segmented_words)):
            expected = [
                m for m in self._rendered_segmented_words[num_test] if m != ""
            ]
            if whitespace:
                predicted_whitespace = subword_tokenizer.encode(
                    self.words[num_test], out_type=str
                )
                predicted = [m.replace("▁", "") for m in predicted_whitespace]
            else:
                predicted = subword_tokenizer.encode(
                    self._rendered_words[num_test], out_type=str
                )
            if expected == predicted:
                test_successful.append(predicted)
            else:
                test_failed.append(predicted)
        """
            The test was successful if all the words were segmented as 
            expected.
        """
        return len(test_successful) == len(self._rendered_segmented_words)

    def solve(self, model: SentencepieceBaseModel, **kwargs) -> bool:
        self._retries_count += 1
        return self._test_run(
            model.sp_preprocessor, kwargs["whitespace_as_prefix"]
        )

    def get_result_info(self) -> Dict:
        info = {"Test info": "Nothing to show yet"}
        return info


class SentencepieceProblemINV(Problem):
    """Class purpose is to conduct 'Invariance Test' for vocabulary.

    Perturbation types offered by this class:
        -substitution.
        -punctuation marks.
        -capitalization.
        -uppercase.
    """

    def __init__(
        self,
        course_id: str,
        assessment_id: str,
        schedule: str,
        problem_id: str,
        problem_type: str,
        parameters: str,
        language_competence: str,
        weight: str,
        retry: str,
        observations: str,
    ):
        """

        Args:
            course_id(str): id of the course to which this problem belongs.
            assessment_id(str): id of the assessment that bundles this problem.
            schedule(str): lesson schedule to execute this problem.
            problem_id(str): id for this problem.
            problem_type(str): problem type enumeration string.
            parameters (str): JSON string parameters for this problem.
            language_competence(str): language competence enumeration value for this problem.
            weight(str): float string value for the weight of this particular problem
                in the overall course score.
            observations(str): observations string. Only descriptive.
        """
        Problem.__init__(
            self,
            course_id,
            assessment_id,
            schedule,
            problem_id,
            problem_type,
            parameters,
            language_competence,
            weight,
            retry,
            observations,
        )
        self._words = self.parameters["words"]
        self._substituted_words: List[Dict] = []
        """
            JSON parameters can specifiy substitutions
        """
        if "substitutions" in self.parameters:
            self._substitution_requested = True
            self._substitution(self.parameters["substitutions"])
        else:
            self._substitution_requested = False

    def _zip_lists(self, list1: list, list2: list):
        zipped_list = list()
        for i in range(len(list2)):
            zipped_list.append([list1[i], list2[i]])
        if list1[-1] != [""]:
            zipped_list.append(list1[-1])
        return zipped_list

    def _check_test(self, expected: list, predicted: list) -> bool:
        return len(expected) == len(predicted)

    def _check_test_punct(self, expected: list, predicted: list) -> bool:
        return len(expected) == len(predicted) - 1

    def _substitution(self, substitutions: dict) -> list:
        """This method takes as an argument a dictionary in which a key
        represents a letter or segment found in one or more of the words
        of the class "Perturb", and the value is a list of letters or
        segments that replace the corresponding key. It searches the words
        to be tested and replaces the letters according to the value of the
        dictionary.

        Args:
            substitutions(dict): dictionary in which each key represents
                a letter or segment found in one or more of the words of
                the class "Perturb".

        Example:
            input: {'a': ['a', 'o']}
            output: [{'niña': [['niñ'], ['a', 'o']]},
                {'lora': [['lor'], ['a', 'o']]},
                {'enfermera': [['enfermer'], ['a', 'o']]},
                {'dueña': [['dueñ'], ['a', 'o']]}]

        Returns:
            Returns a list of dictionaries.
        """
        # not working when a substitution is not in all words.
        ret_words = list()
        for s in substitutions.items():
            source = s[0]
            target = s[1]
            for w in self._words:
                constant = [[m] for m in w.split(source)]
                if constant[0] == "":
                    constant[0] = source
                substitutions = [target]
                ret_word = self._zip_lists(
                    constant, (substitutions * (len(constant) - 1))
                )
                ret_words.append({w: [j for i in ret_word for j in i]})
        self._substituted_words = ret_words
        return ret_words

    def _substitution_render_words(self) -> list:
        """This method renders the words as defined in the
        'substitution method'.

        Example:
            output: [{'niña': ['niña', 'niño']},
                {'lora': ['lora', 'loro']},
                {'enfermera': ['enfermera', 'enfermero']},
                {'dueña': ['dueña', 'dueño']}]

        Returns:
            Returns a list of words.
        """
        render_words = list()
        for word in self._substituted_words:
            w = list(word.values())[0]
            original = list(word.keys())[0]
            words = [m for m in itertools.product(*w)]
            render_word = ["".join(word) for word in words]
            render_words.append({original: render_word})
        return render_words

    def _test_substitution(self, subword_tokenizer) -> bool:
        """This method returns the results of the subword_tokenizer,
        for the words in which a substitution disturbance has been
        made through the 'substitution method'.

        Args:
            subword_tokenizer: an object that represents a tokenization model.

        Returns:
            Returns the results of the subword_tokenizer.
        """
        ret_words = self._substitution_render_words()
        if not self._substitution_requested:
            return True
        else:
            return self._test_run(
                ret_words, subword_tokenizer, self._check_test
            )

    def _punctuation_marks(self):
        """This method adds punctuation marks to the words rendered with
        the 'substitution_render_words method'.

        Example:
            output: [{'niña': ['niña?', 'niña!', 'niña,', 'niña.', 'niña;', 'niña:', 'niña-']},
                {'lora': ['lora?', 'lora!', 'lora,', 'lora.', 'lora;', 'lora:', 'lora-']},
                {'enfermera': ['enfermera?',
                'enfermera!',
                'enfermera,',
                'enfermera.',
                'enfermera;',
                'enfermera:',
                'enfermera-']},
                {'dueña': ['dueña?',
                'dueña!',
                'dueña,',
                'dueña.',
                'dueña;',
                'dueña:',
                'dueña-']}]

        Returns:
            Returns a list of words.
        """
        ret_words = list()
        marks = ["?", "!", ",", ".", ";", ":", "-"]
        for w in self._words:
            ret_words.append({w: [w + m for m in marks]})
        return ret_words

    def _test_punctuation_marks(self, subword_tokenizer):
        """This method returns the results of the subword_tokenizer,
        for the words in which punctuation_marks have been
        added through the 'punctuation_marks method'.

        Args:
            subword_tokenizer: an object that represents a tokenization model.

        Returns:
            Returns the results of the subword_tokenizer.
        """
        ret_words = self._punctuation_marks()
        return self._test_run(
            ret_words, subword_tokenizer, self._check_test_punct
        )

    def _uppercase(self):
        """This method changes words to uppercase.

        Example:
            output: [{'bailaba': ['BAILABA']}, {'tronaba':
            ['TRONABA']}, {'gastaba': ['GASTABA']}]

        Returns:
            Returns a list of words.
        """
        return [{w: [w.upper()]} for w in self._words]

    def _test_uppercase(self, subword_tokenizer):
        """This method returns the results of the subword_tokenizer,
        for the words that were changed to uppercase through the
        'uppercase method'.

        Args:
            subword_tokenizer: an object that represents a tokenization model.

        Returns:
            Returns the results of the subword_tokenizer.
        """
        ret_words = self._uppercase()
        return self._test_run(ret_words, subword_tokenizer, self._check_test)

    def _capitals(self):
        """This method capitalizes words.

        Example:
            output: [{'niña': ['Niña']},
                {'lora': ['Lora']},
                {'enfermera': ['Enfermera']},
                {'dueña': ['Dueña']}]

        Returns:
            Returns a list of words.
        """
        return [{w: [w.capitalize()]} for w in self.words]

    def _test_capitals(self, subword_tokenizer):
        """This method returns the results of the subword_tokenizer,
        for the words which were capitalize through the 'capitals
        method'.

        Args:
            subword_tokenizer: an object that represents a tokenization model.

        Returns:
            Returns the results of the subword_tokenizer.
        """
        ret_words = self.capitals()
        return self._test_run(ret_words, subword_tokenizer, self._check_test)

    def _test_run(self, ret_words: list, subword_tokenizer, check) -> bool:
        test_successful = list()
        test_failed = list()
        total_tested_cases: int = 0
        for w in ret_words:
            for k, v in w.items():
                for num_test in range(len(v)):
                    expected = subword_tokenizer.encode(k, out_type=str)
                    predicted = subword_tokenizer.encode(
                        v[num_test], out_type=str
                    )
                    total_tested_cases += 1
                    if check(expected, predicted):
                        test_successful.append(predicted)
                    else:
                        test_failed.append(predicted)
        return len(test_successful) == total_tested_cases

    def solve(self, model: SentencepieceBaseModel, **kwargs) -> bool:
        self._retries_count += 1
        substitution_solved = self._test_substitution(model.sp_preprocessor)
        uppercase_solver = self._test_uppercase(model.sp_preprocessor)
        capitals_solved = self._test_capitals(model.sp_preprocessor)
        punctuations_solved = self._test_punctuation_marks(
            model.sp_preprocessor
        )
        return (
            substitution_solved
            and uppercase_solver
            and capitals_solved
            and punctuations_solved
        )

    def get_result_info(self) -> Dict:
        info = {"Test info": "Nothing to show yet"}
        return info


def MFT_vocab_test(test_cases: list, subword_tokenizer):
    def _check_test(expected: list, predicted: list) -> bool:
        return expected == predicted

    tests = list()

    for t in test_cases:
        expected = t
        predicted = subword_tokenizer.encode("".join(t), out_type=str)
        tests.append(
            {
                "expected": expected,
                "predicted": predicted,
                "pass": _check_test(expected, predicted),
            }
        )

    return tests


class TestReport:
    """Class purpose is to create an object that allows you to add 
    vocabulary tests as rows of a Pandas' Dataframe.

    This class is deprecated. Compatibility has been broken with new
    problem classes. It might be useful to give ideas for the curriculum
    report mechanism. Eventually will be removed.
    """

    results = pd.DataFrame()

    def add_test(self, test: list, label: str = None, test_type: str = None):
        """This method allows to add the tests to the object 
        created with TestReport(). Each time the add_test() 
        method is executed, a row is added to the Dataframe.

        Args:
            test(list): is the output of the classes.
            label(str): Optional strings shown in each row of 
                the table. None value by default. 
            test_type(str): Optional strings shown in each row 
                of the table. None value by default.       
        """
        df = pd.DataFrame(test)
        label = label
        success = df["pass"].sum()
        fail = (~df["pass"]).sum()
        total = len(df)
        success_ratio = success / total * 100
        examples = list(df[~df["pass"]].predicted[:7])
        columns = [
            "test_type",
            "label",
            "successful",
            "failed",
            "total tests",
            "% pass",
            "examples of failed",
        ]
        data = [
            [test_type, label, success, fail, total, success_ratio, examples]
        ]
        test_df = pd.DataFrame(data, columns=columns)
        self.results = self.results.append(test_df).reset_index(drop=True)

    def print_results(self):
        """Prints the results from the 'add_test method'.
        """
        print(self.results, sep="\n")



from mammut.curriculum.lesson import lesson_base
from mammut.curriculum.models.hugging_face.bert import *
from mammut.curriculum.models.hugging_face.roberta import (
    HuggingFaceRobertaTokenClassificationModel,
    HuggingFaceRobertaSequenceClassificationModel,
    HuggingFaceRobertaQuestionAnsweringModel,
)
from typing import Optional
from mammut.common.corpus.corpus_map import CorpusMap
from mammut.curriculum.core.mammut_session_context import MammutSessionContext


class HuggingFaceBertBaseLesson(lesson_base.Lesson):
    """Implements a concrete Lesson to work with bare Bert
    HuggingFace model curriculum wrapper.
    """

    def __init__(
        self,
        lesson_id: str,
        course: str,
        order: str,
        corpus_id: str,
        corpus_map: CorpusMap,
        name: str,
        language_competence: str,
        parameters: str,
        mammut_session_context: MammutSessionContext,
    ):
        """
        Args:
            lesson_id(str): 'id' value in datasource retrieved data.
            course(str): 'course' value in datasource retrieved data.
            order(str): 'order' value in datasource retrieved data.
            corpus_id(str): 'corpus_id' value in datasource retrieved data.
            name(str): 'name' value in datasource retrieved data.
            language_competence(str): 'language_competence' value in datasource retrieved data.
            mammut_session_context: The session context from the classroom.
        """
        super(HuggingFaceBertBaseLesson, self).__init__(
            lesson_id,
            course,
            order,
            corpus_id,
            corpus_map,
            name,
            language_competence,
            parameters,
            mammut_session_context,
        )

    def _get_model_instance(self) -> HuggingFaceBertBaseModel:
        return HuggingFaceBertBaseModel(
            self.parameters_dict,
            self.course,
            self.lesson_id,
            self._mammut_session_context,
            self.corpus_id,
            self._corpus_map_reference,
        )

    def lecture(self):
        """Loads and prepare the model and tokenizer for this lesson.

        Currently, this data is prepared:
            - Download the pretrained model in memory.
        """
        self._model.load_pretrained_models()

    def practice(self, mammut_session_context: MammutSessionContext):
        """Practice the lessons by training the model with prepared data.

        Todo: There is nothing to train right now. Collaboration between
            corpus (or another data sources) and the model must be defined.

        Args:
            mammut_session_context: The session context from the classroom.
        """
        self._model.save(mammut_session_context)


class HuggingFaceBertTokenClassificationLesson(lesson_base.Lesson):
    """Implements a concrete Lesson to work with the HuggingFaceBertTokenClassificationModel.
    """

    def __init__(
        self,
        lesson_id: str,
        course: str,
        order: str,
        corpus_id: str,
        corpus_map: CorpusMap,
        name: str,
        language_competence: str,
        parameters: str,
        mammut_session_context: MammutSessionContext,
    ):
        """
        Args:
            lesson_id(str): 'id' value in datasource retrieved data.
            course(str): 'course' value in datasource retrieved data.
            order(str): 'order' value in datasource retrieved data.
            corpus_id(str): 'corpus_id' value in datasource retrieved data.
            name(str): 'name' value in datasource retrieved data.
            language_competence(str): 'language_competence' value in datasource retrieved data.
            mammut_session_context: The session context from the classroom.
        """
        super(HuggingFaceBertTokenClassificationLesson, self).__init__(
            lesson_id,
            course,
            order,
            corpus_id,
            corpus_map,
            name,
            language_competence,
            parameters,
            mammut_session_context,
        )

    def _get_model_instance(self) -> HuggingFaceBertTokenClassificationModel:
        return HuggingFaceBertTokenClassificationModel(
            self.parameters_dict,
            self.course,
            self.lesson_id,
            self._mammut_session_context,
            self.corpus_id,
            self._corpus_map_reference,
        )

    def lecture(self):
        """Loads and prepare the model and tokenizer for this lesson.

        Currently, this data is prepared:
            - Download the pretrained model in memory.
        """
        self._model.load_pretrained_models()

    def practice(self, mammut_session_context: MammutSessionContext):
        """Practice the lessons by training the model with prepared data.

        Todo: There is nothing to train right now. Collaboration between
            corpus (or another data sources) and the model must be defined.

        Args:
            mammut_session_context: The session context from the classroom.
        """
        self._model.save(mammut_session_context)


class HuggingFaceRobertaTokenClassificationLesson(lesson_base.Lesson):
    """Implements a concrete Lesson to work with bare Roberta
    HuggingFace model curriculum wrapper.
    """

    def __init__(
        self,
        lesson_id: str,
        course: str,
        order: str,
        corpus_id: str,
        corpus_map: CorpusMap,
        name: str,
        language_competence: str,
        parameters: str,
        mammut_session_context: MammutSessionContext,
    ):
        """
        Args:
            lesson_id(str): 'id' value in datasource retrieved data.
            course(str): 'course' value in datasource retrieved data.
            order(str): 'order' value in datasource retrieved data.
            corpus_id(str): 'corpus_id' value in datasource retrieved data.
            name(str): 'name' value in datasource retrieved data.
            language_competence(str): 'language_competence' value in datasource retrieved data.
            mammut_session_context: The session context from the classroom.
        """
        super(HuggingFaceRobertaTokenClassificationLesson, self).__init__(
            lesson_id,
            course,
            order,
            corpus_id,
            corpus_map,
            name,
            language_competence,
            parameters,
            mammut_session_context,
        )
        self.dataset = self.get_corpus(self.corpus_id, self._corpus_map_reference)

    def get_corpus(
        self, corpus_id: int, corpus_map: CorpusMap,
    ):
        """Returns a training corpus from the corpus map.

        Args:
        corpus_id(int): corpus id in corpus map sheet
            in package.
        corpus_map: CorpusMap object from within the
            Package used.

        Returns:
        The corpus instance.
        """

        corpus = corpus_map.get_corpus_by_id(corpus_id)
        dataset = corpus.get_hf_dataset()
        return dataset

    def _get_model_instance(
        self,
    ) -> HuggingFaceRobertaTokenClassificationModel:
        return HuggingFaceRobertaTokenClassificationModel(
            self.parameters_dict,
            self.course,
            self.lesson_id,
            self._mammut_session_context,
            self.corpus_id,
            self._corpus_map_reference,
        )

    def lecture(self):
        """Loads and prepare the model and tokenizer for this lesson.

        Currently, this data is prepared:
            - Download the pretrained model in memory.
        """

        self.model.load_pretrained_models(self.dataset)

    def practice(self, mammut_session_context: MammutSessionContext):
        """Practice the lessons by training the model with prepared data.

        Args:
        mammut_session_context:  The session context from the classroom.
        """
        tokenized_data = self.model.tokenize_corpus(self.dataset)
        self.model.train(mammut_session_context, **{"corpus": tokenized_data})
        self.model.save(mammut_session_context)


class HuggingFaceRobertaSequenceClassificationLesson(lesson_base.Lesson):
    """Implements a concrete Lesson to work with bare Roberta
    HuggingFace model curriculum wrapper.
    """

    def __init__(
        self,
        lesson_id: str,
        course: str,
        order: str,
        corpus_id: str,
        corpus_map: CorpusMap,
        name: str,
        language_competence: str,
        parameters: str,
        mammut_session_context: MammutSessionContext,
    ):
        """
        Args:
            lesson_id(str): 'id' value in datasource retrieved data.
            course(str): 'course' value in datasource retrieved data.
            order(str): 'order' value in datasource retrieved data.
            corpus_id(str): 'corpus_id' value in datasource retrieved data.
            name(str): 'name' value in datasource retrieved data.
            language_competence(str): 'language_competence' value in datasource retrieved data.
            mammut_session_context: The session context from the classroom.
        """
        super(HuggingFaceRobertaSequenceClassificationLesson, self).__init__(
            lesson_id,
            course,
            order,
            corpus_id,
            corpus_map,
            name,
            language_competence,
            parameters,
            mammut_session_context,
        )
        self.preprocessed_data = None
        self._dataset = None

    def _get_model_instance(
        self,
    ) -> HuggingFaceRobertaSequenceClassificationModel:
        return HuggingFaceRobertaSequenceClassificationModel(
            self.parameters_dict,
            self.course,
            self.lesson_id,
            self._mammut_session_context,
            self.corpus_id,
            self._corpus_map_reference,
        )

    def lecture(self):
        """Loads and prepare the model, tokenizer and data for this lesson.

        Currently, this data is prepared:
            - Download the pretrained model in memory.
        """
        self.model.load_pretrained_models()

        corpus = self._corpus_map_reference.get_corpus_by_id(self.corpus_id)
        self._dataset = corpus.get_hf_dataset()

        self.preprocessed_data = self.model.tokenize_corpus(self._dataset)

    def practice(self, mammut_session_context: MammutSessionContext):
        """Practice the lessons by training the model with prepared data.

        Args:
            mammut_session_context: The session context from the classroom.
        """
        self.model.train(
            mammut_session_context, **{"corpus": self.preprocessed_data}
        )
        self.model.save(mammut_session_context)


class HuggingFaceRobertaQuestionAnsweringLesson(lesson_base.Lesson):
    """Implements a concrete Lesson to work with bare Roberta
    HuggingFace model curriculum wrapper.
    """

    def __init__(
        self,
        lesson_id: str,
        course: str,
        order: str,
        corpus_id: str,
        corpus_map: CorpusMap,
        name: str,
        language_competence: str,
        parameters: str,
        mammut_session_context: MammutSessionContext,
    ):
        """
        Args:
            lesson_id(str): 'id' value in datasource retrieved data.
            course(str): 'course' value in datasource retrieved data.
            order(str): 'order' value in datasource retrieved data.
            corpus_id(str): 'corpus_id' value in datasource retrieved data.
            name(str): 'name' value in datasource retrieved data.
            language_competence(str): 'language_competence' value in datasource retrieved data.
            mammut_session_context: The session context from the classroom.
        """
        super(HuggingFaceRobertaQuestionAnsweringLesson, self).__init__(
            lesson_id,
            course,
            order,
            corpus_id,
            corpus_map,
            name,
            language_competence,
            parameters,
            mammut_session_context,
        )
        self.preprocessed_data = None
        self._dataset = None

    def _get_model_instance(self) -> HuggingFaceRobertaQuestionAnsweringModel:
        return HuggingFaceRobertaQuestionAnsweringModel(
            self.parameters_dict,
            self.course,
            self.lesson_id,
            self._mammut_session_context,
            self.corpus_id,
            self._corpus_map_reference,
        )

    def lecture(self):
        """Loads and prepare the model, tokenizer and data for this lesson.

        Currently, this data is prepared:
            - Download the pretrained model in memory.
        """
        self.model.load_pretrained_models()
        corpus = self._corpus_map_reference.get_corpus_by_id(self.corpus_id)
        self._dataset = corpus.get_hf_dataset()

        self.preprocessed_data = self.model.tokenize_corpus(self._dataset)

        return self.preprocessed_data

    def practice(self, mammut_session_context: MammutSessionContext):
        """Practice the lessons by training the model with prepared data.

        Args:
        mammut_session_context: The session context from the classroom.
        """
        self.model.train(mammut_session_context, **{"corpus": self.preprocessed_data})
        self.model.save(mammut_session_context)



from abc import ABC, abstractmethod
from typing import List


class TokenGenerator(ABC):
    """Abstract class to model a List of tokens generator.

    Tokens are instances of a word in a text.
    A token can be:
        - a Lemma
        - inflected forms of a lemma
        - functional words
        - semantic primes

    """

    @abstractmethod
    def get_tokens(self, pos: str = "", regional_settings: str = "es") -> List:
        """ Get a list of tokens given the implementation.

        Args:
            regional_settings(str): regional settings of tokens retrieved.
            pos(str): POS tag if required.
        """
        pass



from mammut.common.corpus.synthetic_corpus import SyntheticCorpus
from mammut.curriculum.models.model_base import ModelBase
import mammut.curriculum.report as report
from transformers import PreTrainedModel
from mammut.common.corpus.corpus_map import CorpusMap
from datasets import *
from mammut.curriculum.core.mammut_session_context import MammutSessionContext
from mammut.curriculum.report.errors.json_parsing_errors import (
    ModelRequiredParameterMissingKey,
)
from typing import Dict


class HuggingFacePreTrainedModel(ModelBase):
    """This is a wrapper for hugging face pretrained models, to provide
    collaboration between hugging face models and currriculum code.

    This provides a `_model` field to store the concrete model instance.
    Child classes are responsibly for correct initialization of `_model`
    field.

    This also provides a `_tokenizer` field to store the tokenizer for the
    model. When pretraining, the tokenizer will be needed. Also, if predictions
    are computed using the model directly, the input needs to be tokenized
    explicitly.

    The Model can easily receive the context from the Lesson as a simple association
    when required.
    """

    # Parameters constants class attributes.
    PRETRAINED_FROM_PARAM_STR = "pretrained_from"

    def __init__(
        self,
        parameters: Dict,
        course_id: int,
        lesson_id: int,
        mammut_session_context: MammutSessionContext,
        corpus_id: int,
        corpus_map: CorpusMap,
    ):
        """
        Args:
           parameters: The model parameters parsed JSON.
           course_id (int): the course ID to which this model belongs.
           lesson_id (int): the lesson ID to which this model belongs.
           mammut_session_context: Mammut session context with general information about
                current curriculum instance.
        """
        super(HuggingFacePreTrainedModel, self).__init__(
            parameters, course_id, lesson_id
        )

        if self.PRETRAINED_FROM_PARAM_STR not in self.parameters:
            raise ModelRequiredParameterMissingKey(
                self.PRETRAINED_FROM_PARAM_STR, course_id, lesson_id
            )

        else:
            pretrained_from = parameters[self.PRETRAINED_FROM_PARAM_STR]
            if isinstance(pretrained_from, str):
                self._name_or_path = pretrained_from

            elif isinstance(pretrained_from, dict):
                if "course" not in pretrained_from:
                    raise ModelRequiredParameterMissingKey(
                        "pretrained_from.course", course_id, lesson_id
                    )

                if "lesson" not in pretrained_from:
                    raise ModelRequiredParameterMissingKey(
                        "pretrained_from.lesson", course_id, lesson_id
                    )

                pretrained_course_id = pretrained_from["course"]
                pretrained_lesson_id = pretrained_from["lesson"]
                self._name_or_path = self.get_model_file_system_path(
                    mammut_session_context.curriculum_save_folder,
                    pretrained_course_id,
                    pretrained_lesson_id,
                )

        self._model: Optional[PreTrainedModel] = None

    def save(
        self, mammut_session_context: MammutSessionContext, **kwargs
    ) -> None:
        """Saves the model in the file system path provided.

        Args:
            save_model_path: file system path to save the trained model.
            **kwargs: Any other custom attribute.
        """
        model_file_system_path = self._get_model_file_system_path(
            mammut_session_context.curriculum_save_folder
        )
        report.send_message(
            report.CurriculumGeneralDebugMessage(
                f"Saving HuggingFacePretrainedModel in {model_file_system_path}"
            )
        )
        self._model.save_pretrained(model_file_system_path)
        self._tokenizer.save_pretrained(model_file_system_path)

    def load_pretrained_models(self) -> None:
        """Load the pretrained Hugging face model.

        All hugging face models can be load from a pre-trained model,
        which could be stored in the remote hugging face's models hub,
        or in the local file system.
        """
        pass

    def get_corpus(
        self, corpus_id: int, corpus_map: CorpusMap,
    ):
        """Returns a training corpus from the corpus map.

        Args:
        corpus_id(int): corpus id in corpus map sheet
            in package.
        corpus_map: CorpusMap object from within the
            Package used.

        Returns:
        The corpus instance.
        """

        corpus = corpus_map.get_corpus_by_id(corpus_id)
        dataset = corpus.get_hf_dataset()
        return dataset



import enum
from mammut.curriculum.lesson.sentencepiece_lessons import (
    SentencepieceBPELesson,
    SentencepieceNGRAMLesson,
)
from mammut.curriculum.lesson.mock_lesson import MockLesson
from mammut.curriculum.lesson.hugging_face_lessons import (
    HuggingFaceBertBaseLesson,
    HuggingFaceBertTokenClassificationLesson,
    HuggingFaceRobertaTokenClassificationLesson,
    HuggingFaceRobertaSequenceClassificationLesson,
    HuggingFaceRobertaQuestionAnsweringLesson
)
from typing import List, Tuple


class LessonType(enum.Enum):
    """Enumeration for Lessons types, based on models tags.

    Enumeration values are custom classes for lessons types, which encapsulates
    lessons models.

    Each LessonType name maps to a model concrete type class.
    """

    SentencePieceBPE = SentencepieceBPELesson
    SentencePieceNGRAM = SentencepieceNGRAMLesson
    TestMock = MockLesson
    HuggingFaceBertBase = HuggingFaceBertBaseLesson
    HuggingFaceBertTokenClassification = (
        HuggingFaceBertTokenClassificationLesson
    )
    HuggingFaceRobertaTokenClassification = HuggingFaceRobertaTokenClassificationLesson
    HuggingFaceRobertaSequenceClassification = HuggingFaceRobertaSequenceClassificationLesson
    HuggingFaceRobertaQuestionAnswering = HuggingFaceRobertaQuestionAnsweringLesson


    @classmethod
    def list(cls) -> List[Tuple]:
        """Returns a list of pairs for the enumeration names and values
        """
        return list(map(lambda e: (e.name, e.value), cls))

    @classmethod
    def names(cls) -> List[str]:
        """Returns a list of pairs for the enumeration names and values
        """
        return list(map(lambda e: e.name, cls))



from mammut.assessment.problem import Problem
from typing import List, Dict
from mammut.curriculum.lesson import lesson_base


class Assessment:
    """Assessment class bundles multiple problems together.

    The main concern of this class is encapsulate in a single
    unit a set of problems.
    """

    def __init__(self, assessment_id, problems: List[Problem], course_id: int):
        self.assessment_id = int(assessment_id)
        self.errors: List[str] = []
        self.valid = True
        self.problems: List[Problem] = problems
        self.course_id = course_id
        self.approved_problems_weights_sum: float = 0.0
        self.approved_problems_id: List = []
        self.failed_problems_id: List = []
        self.failed_problems_info: Dict = {}

    def apply(self, lesson: lesson_base.Lesson) -> float:
        """This method applies the evaluation process
        encapsulated on each problem, to a trained model
        contained by a Lesson.

        Returns:
            Returns the weights sum for the approved problems.
        """
        self.approved_problems_weights_sum: float = 0.0
        for p in self.problems:
            solved = False
            while not solved and p.is_retry_allowed():
                solved = p.solve(lesson.model)

            if solved:
                self.approved_problems_weights_sum += p.weight
                self.approved_problems_id.append(p.problem_id)
            else:
                self.failed_problems_id.append(p.problem_id)
                p_results = p.get_result_info()
                p_results["retries"] = p.retries_count
                self.failed_problems_info[p.problem_id] = p_results

        return self.approved_problems_weights_sum



import re
from mammut.common.lexicon.inflection_rule import InflectionRule
from mammut.common.storage.storage_manager import StorageManager


class ParadigmProducer:
    REUSE_PARADIGM_CMD_NAME_REGEX = re.compile(
        f"^<({InflectionRule.PARADIGM_NAME_REGEX_STR}\:)?{InflectionRule.PARADIGM_NAME_REGEX_STR}>$"
    )

    def __init__(
        self, word_model: str, grammar: str, hidden: bool, descriptions: {}, group
    ):
        from mammut.common.lexicon.pos import POSdescriptor
        self.word_model = word_model
        self.grammar = grammar
        self.hidden = hidden
        self.descriptions = descriptions
        self.group = group
        self.inflection_rules = []
        for rr in self.grammar.split("|"):
            rrs = rr.strip()
            if ParadigmProducer.REUSE_PARADIGM_CMD_NAME_REGEX.match(rrs) != None:
                name_parts = rrs.strip("<>").split(":")
                word_model_to_reuse = ""
                group_to_look = None
                if len(name_parts) == 1:
                    word_model_to_reuse = name_parts[0]
                    group_to_look = self.group
                else:
                    word_model_to_reuse = name_parts[1]
                    group_to_look = POSdescriptor.SHEETS_LOADED[name_parts[0]]
                rules_to_reuse = group_to_look.producers[
                    word_model_to_reuse
                ].inflection_rules
                self.inflection_rules = self.inflection_rules + rules_to_reuse
            else:
                self.inflection_rules.append(InflectionRule(rrs, self))

    def apply(self, lemma: str) -> []:
        res = []
        for r in self.inflection_rules:
            res.append(r.apply(lemma))
        return res


class ParadigmProducerGroup:
    DEFAULT_START_CELL = "A1"
    DEFAULT_START_COLUMN = "A"
    DEFAULT_START_ROW = "1"
    DEFAULT_END_CELL = "L503"
    DEFAULT_END_COLUMN = "L"
    storage_manager = StorageManager()
    WORD_MODEL_COLUMN_NAME = "word-model"
    GRAMMAR_COLUMN_NAME = "grammar"
    HIDDEN_COLUMN_NAME = "hidden"
    DESCRIPTION_COLUMN_NAME_PREFIX = "description-"

    def __init__(self, sheet_title: str, standard, regional_setting):
        self.sheet_title = sheet_title
        self.standard = standard
        self.regional_setting = regional_setting
        self.producers = {}
        self.feature_classes = []
        if self.sheet_title:
            self.range_name = (
                self.sheet_title + "!" + ParadigmProducerGroup.DEFAULT_START_CELL
            )
            self.data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
                self.standard.spreadsheet_id,
                self.sheet_title,
                self.DEFAULT_START_COLUMN,
                self.DEFAULT_START_ROW,
                self.DEFAULT_END_COLUMN,
            )
            if self.data_frame is not None:
                desc_col_names = []
                for dc in self.data_frame.keys():
                    if dc.startswith(
                        ParadigmProducerGroup.DESCRIPTION_COLUMN_NAME_PREFIX
                    ):
                        desc_col_names.append(dc)
                        desc_name = dc[
                            len(ParadigmProducerGroup.DESCRIPTION_COLUMN_NAME_PREFIX):
                        ]
                        self.feature_classes.append(desc_name)
                for r in self.data_frame.iterrows():
                    descriptions = {}
                    for dc in desc_col_names:
                        desc_name = dc[
                            len(ParadigmProducerGroup.DESCRIPTION_COLUMN_NAME_PREFIX):
                        ]
                        descriptions[desc_name] = r[1][dc]
                    par_prod = ParadigmProducer(
                        r[1][ParadigmProducerGroup.WORD_MODEL_COLUMN_NAME],
                        r[1][ParadigmProducerGroup.GRAMMAR_COLUMN_NAME],
                        r[1][ParadigmProducerGroup.HIDDEN_COLUMN_NAME],
                        descriptions,
                        self,
                    )
                    self.producers[par_prod.word_model] = par_prod

    def get_producers(self):
        return filter(lambda p: not p.hidden, self.producers.keys())



# coding=utf-8

import pandas as pd
import mammut
import logging
from mammut.common.storage.storage_base import SheetsStorageBase
from openpyxl import load_workbook

logger = logging.getLogger()


class EmbeddedStorageExcel(SheetsStorageBase):

    DEFAULT_TOKEN_COLUMN_SEPARATOR = ":"
    DEFAULT_ENGINE = "openpyxl"
    DEFAULT_MAX_ROW = 940

    def __init__(self, storage_path):
        SheetsStorageBase.__init__(self, storage_path)

    def get_spreadsheet_as_dataframe(
        self,
        spreadsheet_id: str,
        sheet_name,
        start_column: str,
        start_row: str,
        end_range: str,
        transpose: bool = False,
    ):
        dataframe = None
        try:
            dataframe = pd.read_excel(
                self.storage_path + spreadsheet_id,
                sheet_name,
                usecols=start_column + self.DEFAULT_TOKEN_COLUMN_SEPARATOR + end_range,
                skiprows=int(start_row) - 1,
                keep_default_na=False,
                dtype=str,
            )
            if transpose:
                dataframe = dataframe.transpose()
        except Exception as ex:
            dataframe = None
            logger.error("Error getting sheet: " + str(ex))
            raise ex

        # This allows to fill sheets in a horizontal way like sheet 'non-functionals' from 'Standard4'
        dataframe.rename(columns=lambda x: x.split(".")[0], inplace=True)
        dataframe = dataframe.groupby(dataframe.columns, axis=1).first()

        return dataframe

    def add_rows_to_spreadsheet(
        self,
        spreadsheet_id: str,
        sheet_name: str,
        rows,
        start_row=SheetsStorageBase.DEFAULT_START_ROW,
        start_col=SheetsStorageBase.DEFAULT_START_COLUMN,
        end_range=SheetsStorageBase.DEFAULT_END_RANGE,
    ):

        try:
            book = load_workbook(self.storage_path + spreadsheet_id)
            ws = book.get_sheet_by_name(sheet_name)

            for i in range(int(start_row), self.DEFAULT_MAX_ROW):
                f = True
                for j in range(0, len(rows[0])):
                    val = ws.cell(row=i, column=j + 1).value
                    if val is not None:
                        f = False
                        break
                if f:
                    for h in range(0, len(rows)):
                        for k in range(0, len(rows[h])):
                            ws.cell(row=i + h, column=k + 1).value = rows[h][k]
                    break

            book.save(self.storage_path + spreadsheet_id)

        except Exception as ex:
            logger.error("Error adding rows on sheet: " + str(ex))
            return None

    def append_row_to_sheet(
        self,
        spreadsheet_id: str,
        sheet_name: str,
        row,
        start_row=SheetsStorageBase.DEFAULT_START_ROW,
        start_col=SheetsStorageBase.DEFAULT_START_COLUMN,
        end_range=SheetsStorageBase.DEFAULT_END_RANGE,
    ):
        try:
            book = load_workbook(self.storage_path + spreadsheet_id)
            ws = book.get_sheet_by_name(sheet_name)

            for i in range(int(start_row), self.DEFAULT_MAX_ROW):
                f = True
                for j in range(0, len(row)):
                    val = ws.cell(row=i, column=j + 1).value
                    if val is not None:
                        f = False
                        break
                if f:
                    for k in range(0, len(row)):
                        ws.cell(row=i, column=k + 1).value = row[k]
                    break

            book.save(self.storage_path + spreadsheet_id)
            return None
        except Exception as ex:
            logger.error("Error adding rows on sheet: " + str(ex))
            return None

    def create_new_sheet(self, spreadsheet_id: str, sheet_name: str):
        book = load_workbook(self.storage_path + spreadsheet_id)
        writer = pd.ExcelWriter(
            self.storage_path + spreadsheet_id, engine=self.DEFAULT_ENGINE
        )
        writer.book = book
        writer.sheets = dict((ws.title, ws) for ws in book.worksheets)
        pd.DataFrame().to_excel(writer, sheet_name)
        writer.save()

    def create_temporal_new_sheet(self, spreadsheet_id: str, base_sheet_title: str):
        sheet_title = (
            base_sheet_title + "-" + mammut.get_str_format_timezone_aware_datetime_now()
        )
        sheet_title = sheet_title.replace(
            mammut.TIME_SEPARATOR, mammut.TIME_ALTERNATIVE_SEPARATOR
        )
        self.create_new_sheet(spreadsheet_id, sheet_title)
        return sheet_title

    def get_temporal_sheets_from_new_to_old(self, spreadsheet_id, base_sheet_title):
        xl = pd.ExcelFile(self.storage_path + spreadsheet_id)
        titles = xl.sheet_names
        result = []
        for title in titles:
            if base_sheet_title in title and len(base_sheet_title) != len(title):
                time_stamp_str = title[len(base_sheet_title) + 1 :]
                time_stamp_str = time_stamp_str.replace(
                    mammut.TIME_ALTERNATIVE_SEPARATOR, mammut.TIME_SEPARATOR
                )
                time_stamp = mammut.str_parse_datetime(time_stamp_str)
                result.append((hash(title + spreadsheet_id), title, time_stamp))
        result = sorted(result, key=lambda i: i[2], reverse=True)
        return result

    def get_sheet_id(self, spreadsheet_id, identifier):
        return hash(identifier + spreadsheet_id)

    def update_row_from_spreadsheet(
        self, spreadsheet_id, sheet_name, row, start_row, start_col
    ):
        book = load_workbook(self.storage_path + spreadsheet_id)
        ws = book.get_sheet_by_name(sheet_name)
        for i in range(0, len(row)):
            ws.cell(row=int(start_row), column=i + 1).value = row[i]
        book.save(self.storage_path + spreadsheet_id)

    def update_rows_from_spreadsheet(
        self, spreadsheet_id, sheet_name, rows, start_row, start_col
    ):
        book = load_workbook(self.storage_path + spreadsheet_id)
        ws = book.get_sheet_by_name(sheet_name)
        for r, row in enumerate(rows):
            for i in range(0, len(row)):
                ws.cell(row=int(start_row) + r, column=i + 1).value = row[i]
        book.save(self.storage_path + spreadsheet_id)

    def delete_rows_from_spreadsheet(
        self, spreadsheet_id, sheet_name, start_index, end_index
    ):
        book = load_workbook(self.storage_path + spreadsheet_id)
        ws = book.get_sheet_by_name(sheet_name)
        ws.delete_rows(int(start_index), int(end_index) - int(start_index) + 1)
        book.save(self.storage_path + spreadsheet_id)

    def delete_sheet(self, spreadsheet_id, sheet_name):
        book = load_workbook(self.storage_path + spreadsheet_id)
        std = book.get_sheet_by_name(sheet_name)
        book.remove_sheet(std)
        book.save(self.storage_path + spreadsheet_id)



# coding=utf-8

import os.path
import tensorflow as tf
import numpy as np

from tensorflow.python.framework import test_util
from tensorflow.python.platform import googletest
from tensorflow.python.platform import tf_logging as logging

from mammut.models.parameter_node import ParameterNode
from mammut.models.worker_node import WorkerNode
from mammut.models.trainable_node import TrainableNode
from mammut.models.trainer import Trainer
from mammut.models import MODELS_CONFIGURATIONS_FILE_NAME


flags = tf.app.flags
FLAGS = flags.FLAGS


# flags.DEFINE_string('test_srcdir', '',
#                     'Dirctory with the source code.')
# flags.DEFINE_string(
#     "test_tmpdir", tf.test.get_temp_dir(), "Temporal Dirctory for data."
# )


class M_Node(ParameterNode):
    def __init__(self):
        ParameterNode.__init__(
            self, "parameter-m", "model-test", "parameter-m", 0, None
        )

    def _operation(self, *args, **kwargs):
        m = tf.Variable(tf.random_uniform([1], -1.0, 1.0), name="m")
        tf.summary.scalar("m", tf.gather(m, 0))
        return m


class MX_Node(WorkerNode):
    def __init__(self):
        WorkerNode.__init__(self, "worker-mx", "model-test", "worker-mx", 0, None)

    def _operation(self, *args, **kwargs):
        m_node = M_Node()
        x = args[0]
        return m_node() * x


class B_Node(ParameterNode):
    def __init__(self):
        ParameterNode.__init__(
            self, "parameter-b", "model-test", "parameter-b", 0, None
        )

    def _operation(self, *args, **kwargs):
        b = tf.Variable(tf.random_uniform([1], -1.0, 1.0), name="b")
        tf.summary.scalar("b", tf.gather(b, 0))

        return b


class LinearRegressorNode(TrainableNode):
    def __init__(self):
        TrainableNode.__init__(
            self, "linear-regresor", "model-test", "linear-regresor", 0, None
        )

    learning_rate = TrainableNode._add_configuration_property(
        "LinearRegressorNode",
        "learning_rate",
        0.5,
        float,
        "Indicates the learning rate of the gradient descent optimizer.",
    )

    def _operation(self, *args, **kwargs):
        x = args[0]
        mx_node = MX_Node()
        b_node = B_Node()

        return mx_node(x) + b_node()

    def _compute_loss(self, resultOp, expected_result):
        loss = tf.reduce_mean(tf.square(expected_result - resultOp))
        tf.summary.scalar("loss", loss)
        return loss

    def _optimize(self, global_step, loss_op):
        optimizer = tf.train.GradientDescentOptimizer(self.learning_rate)
        return [optimizer.minimize(loss_op, global_step), global_step]


class LinearRegressionTrainer(Trainer):
    def __init__(self):
        Trainer.__init__(self, "model-test", "linear-regresion-trainer", None)

    dataset_size = Trainer._add_configuration_property(
        "LinearRegressionTrainer",
        "dataset_size",
        2000,
        int,
        "Indicates the amounts of points to generate dataset.",
    )
    epoch_size = Trainer._add_configuration_property(
        "LinearRegressionTrainer",
        "epoch_size",
        200,
        int,
        "Indicates the amounts of points of the dataset used in each training epoch.",
    )
    step_size = Trainer._add_configuration_property(
        "LinearRegressionTrainer",
        "step_size",
        20,
        int,
        "Indicates the amounts of points of the dataset used in each training step.",
    )
    start_point = Trainer._add_configuration_property(
        "LinearRegressionTrainer",
        "start_point",
        10.1,
        float,
        "Indicates the start X point of the dataset.",
    )
    stop_point = Trainer._add_configuration_property(
        "LinearRegressionTrainer",
        "stop_point",
        200.1,
        float,
        "Indicates the end X point of the dataset.",
    )
    seed = Trainer._add_configuration_property(
        "LinearRegressionTrainer",
        "seed",
        1234,
        int,
        "Indicates the global seed to use.",
    )
    m = Trainer._add_configuration_property(
        "LinearRegressionTrainer",
        "m",
        60.5,
        float,
        "Indicates the slope used to compute Y point of the dataset.",
    )
    b = Trainer._add_configuration_property(
        "LinearRegressionTrainer",
        "b",
        30.2,
        float,
        "Indicates the Y axis intersection point used to compute Y point of the dataset.",
    )

    def _initialize_graph(self):
        self.current_dataset_point = 0
        self.current_epoch = 0
        # self.dataset_input_column = tf.linspace(self.start_point, self.stop_point, self.dataset_size, name="dataset_input_column")
        # self.dataset_input_column = np.linspace(self.start_point, self.stop_point, self.dataset_size, dtype=np.float32)
        # tf.set_random_seed(self.seed)
        np.random.seed(self.seed)
        # noise = tf.random_uniform([self.dataset_size], minval=0.0, maxval=0.5, name='noise')
        # noise = np.random.uniform(0.0, 0.5, self.dataset_size)
        # self.dataset_expected_result_column = ((self.m * self.dataset_input_column) + self.b) + noise
        # self.dataset_input_column = tf.random_uniform([self.dataset_size], self.start_point, self.stop_point, name="dataset_input_column")
        self.dataset_input_column = np.random.rand(self.dataset_size).astype(np.float32)
        self.dataset_expected_result_column = (
            self.dataset_input_column * self.m + self.b
        )

    def _recover_variables(self, sess):
        # self.dataset_input_column, self.dataset_expected_result_column = sess.run([self.dataset_input_column, self.dataset_expected_result_column])
        self.current_dataset_point = 0
        self.current_epoch = 0

    def _build_model(self) -> TrainableNode:
        return LinearRegressorNode()

    def _get_input_placeholder(self):
        return tf.placeholder(tf.float32, shape=(self.step_size), name="X")

    def _get_expectedResult_placeholder(self):
        return tf.placeholder(tf.float32, shape=(self.step_size), name="Y")

    def _fill_feed_dict(self, sess, input_placeholder, expected_result_placeholder):
        feed_dict = {
            input_placeholder: self.dataset_input_column[
                self.current_dataset_point : self.current_dataset_point + self.step_size
            ],
            expected_result_placeholder: self.dataset_expected_result_column[
                self.current_dataset_point : self.current_dataset_point + self.step_size
            ],
        }
        self.current_dataset_point = self.current_dataset_point + self.step_size
        if self.current_dataset_point % self.epoch_size == 0:
            self.current_epoch = self.current_epoch + 1
        return feed_dict

    def _get_current_epoch(self):
        return self.current_epoch


class ModelTest(test_util.TensorFlowTestCase):
    def setUp(self):
        # Creates a task context with the correct testing paths.
        MODELS_CONFIGURATIONS_FILE_NAME = os.path.join(FLAGS.test_tmpdir, "models.ini")

    def testLinearRegresor(self):
        """Test a simple linear regression model."""
        logging.info("Starting test...")
        trainer = LinearRegressionTrainer()
        final_loss = trainer.train(
            "tensorboard",
            int(trainer.dataset_size / trainer.epoch_size),
            log_step_mod=10,
        )
        logging.info("Final loss = {:-f}".format(final_loss))
        self.assertLessEqual(final_loss, 1.0, "Final loss greather than 1.")


if __name__ == "__main__":
    googletest.main()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: apps/Protos.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='apps/Protos.proto',
  package='com.mammut.communication.transducers',
  syntax='proto3',
  serialized_pb=_b('\n\x11\x61pps/Protos.proto\x12$com.mammut.communication.transducers\"\xcb\x01\n\x0f\x43oncreteThought\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x0f\n\x07user_id\x18\x02 \x01(\x04\x12\x11\n\tuser_type\x18\x03 \x01(\t\x12\x13\n\x0buser_gender\x18\x04 \x01(\t\x12\x16\n\x0euser_birthdate\x18\x05 \x01(\t\x12\x12\n\nmachine_id\x18\x06 \x01(\x04\x12\x0f\n\x07room_id\x18\x07 \x01(\x04\x12\x11\n\troom_type\x18\x08 \x01(\t\x12\x0c\n\x04text\x18\t \x01(\x0c\x12\x15\n\rcreation_date\x18\n \x01(\t\"\x91\x06\n\x05Issue\x12\n\n\x02id\x18\x01 \x01(\x04\x12\x13\n\x0b\x63ustom_code\x18\x02 \x01(\t\x12\x1b\n\x13\x63\x61usative_vertex_id\x18\x03 \x01(\x04\x12\x12\n\nmachine_id\x18\x04 \x01(\x04\x12\x0f\n\x07room_id\x18\x05 \x01(\x04\x12\x12\n\nissue_type\x18\x06 \x01(\t\x12\x13\n\x0bissue_state\x18\x07 \x01(\t\x12\x17\n\x0fresolution_text\x18\x08 \x01(\x0c\x12\x15\n\rcreation_date\x18\t \x01(\t\x12\x17\n\x0fresolution_date\x18\n \x01(\t\x12[\n\x12\x63onversation_scope\x18\x0b \x01(\x0b\x32?.com.mammut.communication.transducers.Issue.ConversationClosure\x12\x19\n\x11triggers_messages\x18\x0c \x03(\t\x1a\xba\x03\n\x13\x43onversationClosure\x12o\n\rsnapshotStack\x18\x01 \x03(\x0b\x32X.com.mammut.communication.transducers.Issue.ConversationClosure.GraphTransverseValuePair\x12z\n\x18snapshotKnownValuesStack\x18\x02 \x03(\x0b\x32X.com.mammut.communication.transducers.Issue.ConversationClosure.GraphTransverseValuePair\x12|\n\x1asnapshotUnknownValuesStack\x18\x03 \x03(\x0b\x32X.com.mammut.communication.transducers.Issue.ConversationClosure.GraphTransverseValuePair\x1a\x38\n\x18GraphTransverseValuePair\x12\r\n\x05steps\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\tB\x08\x42\x06Protosb\x06proto3')
)




_CONCRETETHOUGHT = _descriptor.Descriptor(
  name='ConcreteThought',
  full_name='com.mammut.communication.transducers.ConcreteThought',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='com.mammut.communication.transducers.ConcreteThought.id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user_id', full_name='com.mammut.communication.transducers.ConcreteThought.user_id', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user_type', full_name='com.mammut.communication.transducers.ConcreteThought.user_type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user_gender', full_name='com.mammut.communication.transducers.ConcreteThought.user_gender', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user_birthdate', full_name='com.mammut.communication.transducers.ConcreteThought.user_birthdate', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='machine_id', full_name='com.mammut.communication.transducers.ConcreteThought.machine_id', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='room_id', full_name='com.mammut.communication.transducers.ConcreteThought.room_id', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='room_type', full_name='com.mammut.communication.transducers.ConcreteThought.room_type', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='text', full_name='com.mammut.communication.transducers.ConcreteThought.text', index=8,
      number=9, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='creation_date', full_name='com.mammut.communication.transducers.ConcreteThought.creation_date', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=60,
  serialized_end=263,
)


_ISSUE_CONVERSATIONCLOSURE_GRAPHTRANSVERSEVALUEPAIR = _descriptor.Descriptor(
  name='GraphTransverseValuePair',
  full_name='com.mammut.communication.transducers.Issue.ConversationClosure.GraphTransverseValuePair',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='steps', full_name='com.mammut.communication.transducers.Issue.ConversationClosure.GraphTransverseValuePair.steps', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='com.mammut.communication.transducers.Issue.ConversationClosure.GraphTransverseValuePair.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=995,
  serialized_end=1051,
)

_ISSUE_CONVERSATIONCLOSURE = _descriptor.Descriptor(
  name='ConversationClosure',
  full_name='com.mammut.communication.transducers.Issue.ConversationClosure',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='snapshotStack', full_name='com.mammut.communication.transducers.Issue.ConversationClosure.snapshotStack', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='snapshotKnownValuesStack', full_name='com.mammut.communication.transducers.Issue.ConversationClosure.snapshotKnownValuesStack', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='snapshotUnknownValuesStack', full_name='com.mammut.communication.transducers.Issue.ConversationClosure.snapshotUnknownValuesStack', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ISSUE_CONVERSATIONCLOSURE_GRAPHTRANSVERSEVALUEPAIR, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=609,
  serialized_end=1051,
)

_ISSUE = _descriptor.Descriptor(
  name='Issue',
  full_name='com.mammut.communication.transducers.Issue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='com.mammut.communication.transducers.Issue.id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='custom_code', full_name='com.mammut.communication.transducers.Issue.custom_code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='causative_vertex_id', full_name='com.mammut.communication.transducers.Issue.causative_vertex_id', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='machine_id', full_name='com.mammut.communication.transducers.Issue.machine_id', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='room_id', full_name='com.mammut.communication.transducers.Issue.room_id', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='issue_type', full_name='com.mammut.communication.transducers.Issue.issue_type', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='issue_state', full_name='com.mammut.communication.transducers.Issue.issue_state', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='resolution_text', full_name='com.mammut.communication.transducers.Issue.resolution_text', index=7,
      number=8, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='creation_date', full_name='com.mammut.communication.transducers.Issue.creation_date', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='resolution_date', full_name='com.mammut.communication.transducers.Issue.resolution_date', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='conversation_scope', full_name='com.mammut.communication.transducers.Issue.conversation_scope', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='triggers_messages', full_name='com.mammut.communication.transducers.Issue.triggers_messages', index=11,
      number=12, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ISSUE_CONVERSATIONCLOSURE, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=266,
  serialized_end=1051,
)

_ISSUE_CONVERSATIONCLOSURE_GRAPHTRANSVERSEVALUEPAIR.containing_type = _ISSUE_CONVERSATIONCLOSURE
_ISSUE_CONVERSATIONCLOSURE.fields_by_name['snapshotStack'].message_type = _ISSUE_CONVERSATIONCLOSURE_GRAPHTRANSVERSEVALUEPAIR
_ISSUE_CONVERSATIONCLOSURE.fields_by_name['snapshotKnownValuesStack'].message_type = _ISSUE_CONVERSATIONCLOSURE_GRAPHTRANSVERSEVALUEPAIR
_ISSUE_CONVERSATIONCLOSURE.fields_by_name['snapshotUnknownValuesStack'].message_type = _ISSUE_CONVERSATIONCLOSURE_GRAPHTRANSVERSEVALUEPAIR
_ISSUE_CONVERSATIONCLOSURE.containing_type = _ISSUE
_ISSUE.fields_by_name['conversation_scope'].message_type = _ISSUE_CONVERSATIONCLOSURE
DESCRIPTOR.message_types_by_name['ConcreteThought'] = _CONCRETETHOUGHT
DESCRIPTOR.message_types_by_name['Issue'] = _ISSUE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ConcreteThought = _reflection.GeneratedProtocolMessageType('ConcreteThought', (_message.Message,), dict(
  DESCRIPTOR = _CONCRETETHOUGHT,
  __module__ = 'apps.Protos_pb2'
  # @@protoc_insertion_point(class_scope:com.mammut.communication.transducers.ConcreteThought)
  ))
_sym_db.RegisterMessage(ConcreteThought)

Issue = _reflection.GeneratedProtocolMessageType('Issue', (_message.Message,), dict(

  ConversationClosure = _reflection.GeneratedProtocolMessageType('ConversationClosure', (_message.Message,), dict(

    GraphTransverseValuePair = _reflection.GeneratedProtocolMessageType('GraphTransverseValuePair', (_message.Message,), dict(
      DESCRIPTOR = _ISSUE_CONVERSATIONCLOSURE_GRAPHTRANSVERSEVALUEPAIR,
      __module__ = 'apps.Protos_pb2'
      # @@protoc_insertion_point(class_scope:com.mammut.communication.transducers.Issue.ConversationClosure.GraphTransverseValuePair)
      ))
    ,
    DESCRIPTOR = _ISSUE_CONVERSATIONCLOSURE,
    __module__ = 'apps.Protos_pb2'
    # @@protoc_insertion_point(class_scope:com.mammut.communication.transducers.Issue.ConversationClosure)
    ))
  ,
  DESCRIPTOR = _ISSUE,
  __module__ = 'apps.Protos_pb2'
  # @@protoc_insertion_point(class_scope:com.mammut.communication.transducers.Issue)
  ))
_sym_db.RegisterMessage(Issue)
_sym_db.RegisterMessage(Issue.ConversationClosure)
_sym_db.RegisterMessage(Issue.ConversationClosure.GraphTransverseValuePair)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('B\006Protos'))
# @@protoc_insertion_point(module_scope)



# coding=utf-8

import tensorflow as tf

from mammut.models.parameter_node import ParameterNode
from mammut.models.worker_node import WorkerNode
from mammut.models.reparameterization import gumbel


class ObservedUtilityEstimatorNode(ParameterNode):
    def __init__(
        self,
        padding,
        state_free_dimension_choices,
        state_size,
        scope_name,
        configuration_path,
        configuration_name,
        task_index,
        redis_connection_pool,
    ):
        ParameterNode.__init__(
            self,
            scope_name,
            configuration_path,
            configuration_name,
            task_index,
            redis_connection_pool,
        )
        self.size = state_size - self.state_size_reduction
        self.free_dimension_choices = state_free_dimension_choices
        # The size of the linear regressor must be the same as the output of the convolution operation
        if padding == "VALID":
            self.linear_regressor_size = state_size - self.size + 1
        else:
            self.linear_regressor_size = state_size

    state_size_reduction = ParameterNode._add_configuration_property(
        "ObservedUtilityEstimatorNode",
        "state_size_reduction",
        1,
        int,
        "The size of the filter is the state_size - state_size_reduction.",
    )

    def _operation(self, *args, **kwargs):
        filters = []
        options_list = []
        for i in range(0, self.free_dimension_choices):
            # This is a filter to use with conv1d that expects a filter / kernel tensor of shape
            # [filter_width, in_channels, out_channels]
            filters.append(
                tf.Variable(
                    tf.truncated_normal(
                        [self.size, 1, 1],
                        mean=0.0,
                        stddev=0.1,
                        dtype=tf.float32,
                        seed=21,
                        name="init-filter",
                    ),
                    name="utility-paramters-convolution-filter-{}".format(i),
                )
            )
            options_list.append(float(i))
        linear_regressor = tf.Variable(
            tf.truncated_normal(
                [self.linear_regressor_size],
                mean=0.0,
                stddev=0.1,
                dtype=tf.float32,
                seed=21,
                name="init-linear-regressor",
            ),
            name="linear-utility-m",
        )
        options = tf.constant(options_list, dtype=tf.float32, name="options")
        temperature = tf.Variable(0.1, name="temperature", trainable=False)
        return [filters, linear_regressor, options, temperature]


class LexicalDiscreteChoiceNode(WorkerNode):
    def __init__(
        self,
        one_hot_output,
        intermediate_node,
        state_free_dimension,
        state_free_dimension_choices,
        state_size,
        scope_name,
        configuration_path,
        configuration_name,
        task_index,
        redis_connection_pool=None,
    ):
        WorkerNode.__init__(
            self,
            scope_name,
            configuration_path,
            configuration_name,
            task_index,
            redis_connection_pool,
        )
        self.use_one_hot_output = one_hot_output
        self.is_intermediate_node = intermediate_node
        self.free_dimension = state_free_dimension
        self.free_dimension_choices = state_free_dimension_choices
        self.state_size = state_size
        self.observed_utility_estimator = ObservedUtilityEstimatorNode(
            self.padding,
            self.free_dimension_choices,
            self.state_size,
            scope_name,
            configuration_path,
            configuration_name,
            task_index,
            redis_connection_pool,
        )
        (
            self.filters,
            self.linear_regressor,
            self.options,
            self.temperature,
        ) = self.observed_utility_estimator()

    padding = ParameterNode._add_configuration_property(
        "LexicalDiscretChoiceNode",
        "padding",
        "SAME",
        None,
        "The padding parameter of the tf.nn.conv2d method.",
    )
    output_size = ParameterNode._add_configuration_property(
        "LexicalDiscretChoiceNode",
        "output_size",
        1,
        int,
        "The amount of options in the output. The options is a ordered vector",
    )

    def _operation(self, *args, **kwargs):
        # Extract the input state and reshaped as one vector
        x = args[0]
        # This is the input tensor of shape [batch, in_width, in_channels] expected for conv1d
        x = tf.to_float(tf.reshape(x, [-1, self.state_size, 1]))

        # Compute the utility of each option
        utility_options = []
        for i in range(0, self.free_dimension_choices):
            # Applies the filter to the input state in order to obtain the utility parameters of the state for the option i
            utility_parameters = tf.nn.conv1d(
                x,
                self.filters[i],
                stride=1,
                padding=self.padding,
                data_format="NWC",
                name="convolution",
            )
            utility_parameters = tf.reshape(
                utility_parameters, [-1, tf.shape(self.linear_regressor)[0]]
            )
            # Compute the observed utility. Its a vector of [batch_size]
            observed_utility = tf.reduce_sum(
                self.linear_regressor * utility_parameters, axis=1
            )
            utility_options.append(observed_utility)
        utility_options_tensor = tf.stack(
            utility_options, axis=1, name="utility_options"
        )

        # Compute the probability of each option
        options_probabilities = tf.nn.softmax(utility_options_tensor, name="logit")

        # Compute the selected option using Gumbel-Max Trick if you are using the output as input of another node
        if self.is_intermediate_node:
            if self.use_one_hot_output:
                output = gumbel.gumbel_softmax(
                    utility_options_tensor, self.temperature, hard=True
                )
            else:
                output = gumbel.gumbel_softmax_option(
                    utility_options_tensor, self.temperature, self.options
                )
        else:
            output = tf.argmax(options_probabilities, 1)
            if self.use_one_hot_output:
                output = tf.one_hot(output, self.free_dimension_choices)
        return [output, options_probabilities, utility_options_tensor]



"""
    Spreadsheet cells coordinates for dataframe reading.
"""
DEFAULT_START_CELL = "A1"
DEFAULT_START_COLUMN = "A"
DEFAULT_START_ROW = "1"
DEFAULT_END_CELL = "J"

"""
    Columns name for spreadsheet specification
"""
COURSE_ID_COLUMN = "course_id"
ASSESSMENT_ID_COLUMN = "assessment_id"
SCHEDULE_COLUMN = "schedule"
PROBLEM_ID_COLUMN = "problem_id"
PROBLEM_TYPE_COLUMN = "problem_type"
PARAMETERS_COLUMN = "parameters"
LANGUAGE_COMPETENCE_COLUMN = "language_competence"
WEIGHT_COLUMN = "weight"
RETRY_COLUMN = "retry"
OBSERVATIONS_COLUMN = "observations"



import json

from mammut.common.lexicon.dictionary import Dictionary
from mammut.common.lexicon.linguistic_standard import LinguisticStandard
from mammut.common.synthetic.synthetic_corpus_generator.morphemes_generator import (
    MorphemesGenerator,
)
from mammut.common.synthetic.synthetic_corpus_generator.synthetic_corpus_generator import (
    SyntheticCorpusGenerator,
)
from mammut.common.corpus.corpus_operations import CorpusOperations
from typing import List


class SyntheticCorpus(CorpusOperations):
    """Synthetic Corpus support in corpus map sheet.

    Parameters for the SyntheticCorpusGeneratorClass
    can be specified using the configuration_json column
    in corpus map sheet.
    """

    def __init__(
        self,
        spreadsheet_id: str,
        id,
        dictionary: Dictionary,
        standard: LinguisticStandard,
        regional_settings: str,
        json_parameters,
    ):
        """

        :param spreadsheet_id: spreadsheet id used for corpus indexing
        :param id:
        :param dictionary:
        :param standard:
        :param regional_settings:
        :param json_parameters:
        """
        self.configuration = json.loads(json_parameters)
        self.id = id
        self.spreadsheet_id = spreadsheet_id
        self.regional_settings = regional_settings
        self.events: List[str] = []
        """
            Length to split the synthetic data into substrings of text,
            representing sentences. 
        """
        self._events_tokens_length: int = self.configuration[
            "events_tokens_length"
        ]
        self.morphemes_generator = MorphemesGenerator(standard, dictionary)
        self.synthetic_corpus_generator = SyntheticCorpusGenerator(
            self.morphemes_generator
        )
        self._generate_synthetic_tokens()

    def _generate_synthetic_tokens(self):
        """
            Generates the synthetic tokens for this corpus events.
        """
        tokens_text: List[
            str
        ] = self.synthetic_corpus_generator.generate_synthetic_corpus(
            self.configuration["nouns_words"],
            self.configuration["adjectives_words"],
            self.configuration["verbs_words"],
            self.configuration["nouns_bound"],
            self.configuration["adjectives_bound"],
            self.configuration["verbs_bound"],
            self.configuration["free_lemma_words"],
            self.configuration["random_ext"],
        )
        """
            Segment the list in segments of the specified _event_tokens_length.
            After that, each segmented list is joined as a string separated by 
            space character.
        """
        segmented_tokens = list(map(
            lambda token: " ".join(token),
            [
                tokens_text[x: x + self._events_tokens_length]
                for x in range(0, len(tokens_text), self._events_tokens_length)
            ]
        ))
        self.events = segmented_tokens

    @property
    def indexable(self) -> bool:
        return False

    @property
    def annotable(self) -> bool:
        return False

    def get_events_as_strings(self) -> List[str]:
        return self.events



from mammut.curriculum.models.model_base import ModelBase
from sentencepiece import SentencePieceTrainer, SentencePieceProcessor
from typing import List
import io
from abc import abstractmethod
from typing import Optional, Dict
import mammut.curriculum.report as report
from mammut.curriculum.core.mammut_session_context import MammutSessionContext


class SentencepieceBaseModel(ModelBase):
    def __init__(self, parameters: Dict, course_id: int, lesson_id: int):
        ModelBase.__init__(self, parameters, course_id, lesson_id)
        """
            Attribute sp_preprocessor contains the SentencePiece 
            preprocessor object developed by third party.
        """
        self.sp_preprocessor: Optional[SentencePieceProcessor] = None
        self._model_io: io.BytesIO = None

    @property
    @abstractmethod
    def model_name(self) -> str:
        pass

    def train(self, mammut_session_context: MammutSessionContext, **kwargs):
        """Train the model this class implements.

        Args:
            **vocabulary_size (int): vocabulary size, default to 8000.
            **whitespace_as_suffix(bool): treat whitespace marker as suffix
                instead of prefix.
        """
        self._model_io = io.BytesIO()
        vocabulary_size = self.parameters["vocabulary_size"]
        whitespace_as_suffix = self.parameters["treat_whitespace_as_suffix"]
        corpus = kwargs["corpus"]
        self._train_subword_unit(vocabulary_size, whitespace_as_suffix, corpus)

    def save(self, mammut_session_context: MammutSessionContext, **kwargs):
        """Persist the model in memory.
        """
        model_file_system_path = self._get_model_file_system_path(
            mammut_session_context.curriculum_save_folder
        )
        report.send_message(
            report.CurriculumGeneralDebugMessage(
                f"Save Sentencepiece model in: {model_file_system_path}"
            )
        )
        pass

    def _train_subword_unit(
        self, vocabulary_sz: int, whitespace_as_suffix: bool, corpus: List[str]
    ):
        # Todo: Study this params and their meaning.
        pad_id = 0
        unk_id = 1
        bos_id = 2
        eos_id = 3
        corpus_iter = corpus.__iter__()
        SentencePieceTrainer.Train(
            sentence_iterator=corpus_iter,
            model_writer=self._model_io,
            vocab_size=vocabulary_sz,
            treat_whitespace_as_suffix=whitespace_as_suffix,
            pad_id=pad_id,
            unk_id=unk_id,
            bos_id=bos_id,
            eos_id=eos_id,
            model_type=self.model_name,
        )
        self._load_subword_model()

    def _load_subword_model(self):
        self.sp_preprocessor = SentencePieceProcessor(
            model_proto=self._model_io.getvalue()
        )


class SentencePieceBPEModel(SentencepieceBaseModel):
    def __init__(self, parameters: Dict, course_id: int, lesson_id: int):
        SentencepieceBaseModel.__init__(self, parameters, course_id, lesson_id)

    @property
    def model_name(self) -> str:
        return "bpe"


class SentencePieceNGRAMModel(SentencepieceBaseModel):
    def __init__(self, parameters: Dict, course_id: int, lesson_id: int):
        SentencepieceBaseModel.__init__(self, parameters, course_id, lesson_id)

    @property
    def model_name(self) -> str:
        return "unigram"



from mammut.curriculum.report.reporters.base_reporter import BaseReporter
from mammut.curriculum.report.messages.message import CurriculumBaseMessage
import logging


class ConsoleReporter(BaseReporter):
    """Custom reporter for standard output message handling.

    TODO: constructor parameters must be replaced by configuration values.
    """

    def __init__(self, logger_name, threshold=logging.INFO):
        super(ConsoleReporter, self).__init__(threshold)
        self._set_custom_logger(logger_name, threshold)

    def send_message(self, message: CurriculumBaseMessage, **kwargs):
        self._logger.log(message.level, str(message))

    def _set_custom_logger(self, name: str, level):
        logger = logging.getLogger(name)
        logger.setLevel(level)
        ch = logging.StreamHandler()
        ch.setLevel(level)
        formatter = logging.Formatter("[%(levelname)s] %(message)s")
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        logger.propagate = False
        self._logger = logger



# coding=utf-8
from collections import defaultdict
import configparser
import os
from pkg_resources import resource_string
import re
from typing import Dict

from mammut.common.storage.storage_base import SheetsStorageBase, SlidesStorageBase
from mammut.common.storage.embedded_storage_excel import EmbeddedStorageExcel
from mammut.common.storage.emdedded_storage_powerpoint import EmbeddedStoragePowerPoint
from mammut.common.storage.google_storage_sheet import GoogleStorageSheet
from mammut.common.storage.google_storage_slide import GoogleStorageSlide


class StorageManager:
    __STORAGE_INFERRED__: Dict[str, bool] = defaultdict(lambda: False)
    __USE_LOCAL_STORAGE__: Dict[str, bool] = defaultdict(lambda: False)

    current_sheets_implementation: SheetsStorageBase = None
    current_slides_implementation: SlidesStorageBase = None

    def __init__(self, storage_path=""):
        self.configurations = configparser.ConfigParser()
        self.configurations.read_string(
            resource_string("mammut.resources", "features.ini").decode("utf-8")
        )
        self.configurations.read(["data/features.ini", "features.ini"])

        if storage_path == "":
            if not os.getenv("LOCAL_STORAGE_BASE_FOLDER") is None:
                self.local_storage_path = os.getenv("LOCAL_STORAGE_BASE_FOLDER")
            else:
                self.local_storage_path = self.configurations["STORAGE"][
                    "local_storage_base_folder"
                ]
        else:
            self.local_storage_path = storage_path

        self.local_storage_regex = self.configurations["STORAGE"]["local_storage_regex"]
        self.google_storage_regex = self.configurations["STORAGE"][
            "google_storage_regex"
        ]

    # Infer which is the right storage for that running notebook.
    def infer_spreadsheet_storage(self, documentId: str):
        if not StorageManager.__STORAGE_INFERRED__[documentId]:
            if re.match(self.local_storage_regex, documentId) is not None:
                self.current_sheets_implementation = EmbeddedStorageExcel(
                    self.local_storage_path
                )
                StorageManager.__STORAGE_INFERRED__[documentId] = True
                StorageManager.__USE_LOCAL_STORAGE__[documentId] = True
                return self.current_sheets_implementation
            elif re.match(self.google_storage_regex, documentId) is not None:
                self.current_sheets_implementation = GoogleStorageSheet()
                StorageManager.__STORAGE_INFERRED__[documentId] = True
                StorageManager.__USE_LOCAL_STORAGE__[documentId] = False
                return self.current_sheets_implementation
            else:
                raise Exception("Unknown package spreadsheets data storage")
        else:
            if StorageManager.__USE_LOCAL_STORAGE__[documentId]:
                if self.current_sheets_implementation is None:
                    self.current_sheets_implementation = EmbeddedStorageExcel(
                        self.local_storage_path
                    )
                return self.current_sheets_implementation
            else:
                if self.current_sheets_implementation is None:
                    self.current_sheets_implementation = GoogleStorageSheet()
                return self.current_sheets_implementation

    def infer_slides_storage(self, documentId: str):
        if not StorageManager.__STORAGE_INFERRED__[documentId]:
            if re.match(self.local_storage_regex, documentId) is not None:
                self.current_slides_implementation = EmbeddedStoragePowerPoint(
                    self.local_storage_path
                )
                StorageManager.__STORAGE_INFERRED__[documentId] = True
                StorageManager.__USE_LOCAL_STORAGE__[documentId] = True
                return self.current_slides_implementation
            elif re.match(self.google_storage_regex, documentId) is not None:
                self.current_slides_implementation = GoogleStorageSlide()
                StorageManager.__STORAGE_INFERRED__[documentId] = True
                StorageManager.__USE_LOCAL_STORAGE__[documentId] = False
                return self.current_slides_implementation
            else:
                raise Exception("Unknown package slides data storage")
        else:
            if StorageManager.__USE_LOCAL_STORAGE__[documentId]:
                if self.current_slides_implementation is None:
                    self.current_slides_implementation = EmbeddedStoragePowerPoint(
                        self.local_storage_path
                    )
                return self.current_slides_implementation
            else:
                if self.current_slides_implementation is None:
                    self.current_slides_implementation = GoogleStorageSlide()
                return self.current_slides_implementation

    @staticmethod
    def __add_local_storage_sheet_file_extension(documentId: str):
        if StorageManager.__USE_LOCAL_STORAGE__[documentId]:
            if documentId.endswith(".xlsx"):
                return documentId
            else:
                return documentId + ".xlsx"
        else:
            return documentId

    @staticmethod
    def __add_local_storage_presentation_file_extension(documentId: str):
        if StorageManager.__USE_LOCAL_STORAGE__[documentId]:
            if documentId.endswith(".pptx"):
                return documentId
            else:
                return documentId + ".pptx"
        else:
            return documentId

    def get_spreadsheet_as_dataframe(
        self,
        spreadsheet_id: str,
        sheet_name,
        start_column: str,
        start_row: str,
        end_range: str,
        transpose: bool = False,
    ):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            spreadsheet_id
        )
        return self.infer_spreadsheet_storage(
            local_spreadsheet_id
        ).get_spreadsheet_as_dataframe(
            local_spreadsheet_id,
            sheet_name,
            start_column,
            start_row,
            end_range,
            transpose,
        )

    def add_rows_to_spreadsheet(
        self,
        spreadsheet_id: str,
        sheet_name: str,
        rows,
        start_row=SheetsStorageBase.DEFAULT_START_ROW,
        start_col=SheetsStorageBase.DEFAULT_START_COLUMN,
        end_range=SheetsStorageBase.DEFAULT_END_RANGE,
    ):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            spreadsheet_id
        )
        return self.infer_spreadsheet_storage(
            local_spreadsheet_id
        ).add_rows_to_spreadsheet(
            local_spreadsheet_id, sheet_name, rows, start_row, start_col, end_range
        )

    def append_row_to_sheet(
        self,
        spreadsheet_id: str,
        sheet_name: str,
        row,
        start_row=SheetsStorageBase.DEFAULT_START_ROW,
        start_col=SheetsStorageBase.DEFAULT_START_COLUMN,
        end_range=SheetsStorageBase.DEFAULT_END_RANGE,
    ):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            spreadsheet_id
        )
        return self.infer_spreadsheet_storage(local_spreadsheet_id).append_row_to_sheet(
            local_spreadsheet_id, sheet_name, row, start_row, start_col, end_range
        )

    def create_new_sheet(self, spreadsheet_id: str, sheet_name: str):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            spreadsheet_id
        )
        return self.infer_spreadsheet_storage(local_spreadsheet_id).create_new_sheet(
            local_spreadsheet_id, sheet_name
        )

    def create_temporal_new_sheet(self, spreadsheet_id: str, base_sheet_title: str):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            spreadsheet_id
        )
        return self.infer_spreadsheet_storage(
            local_spreadsheet_id
        ).create_temporal_new_sheet(local_spreadsheet_id, base_sheet_title)

    def get_temporal_sheets_from_new_to_old(self, spreadsheet_id, base_sheet_title):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            spreadsheet_id
        )
        return self.infer_spreadsheet_storage(
            local_spreadsheet_id
        ).get_temporal_sheets_from_new_to_old(local_spreadsheet_id, base_sheet_title)

    def get_sheet_id(self, spreadsheet_id, identifier):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            spreadsheet_id
        )
        return self.infer_spreadsheet_storage(local_spreadsheet_id).get_sheet_id(
            local_spreadsheet_id, identifier
        )

    def update_row_from_spreadsheet(
        self, spreadsheet_id, sheet_name, row, start_row, start_col
    ):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            spreadsheet_id
        )
        return self.infer_spreadsheet_storage(
            local_spreadsheet_id
        ).update_row_from_spreadsheet(
            local_spreadsheet_id, sheet_name, row, start_row, start_col
        )

    def update_rows_from_spreadsheet(
        self, spreadsheet_id, sheet_name, rows, start_row, start_col
    ):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            spreadsheet_id
        )
        return self.infer_spreadsheet_storage(
            local_spreadsheet_id
        ).update_rows_from_spreadsheet(
            local_spreadsheet_id, sheet_name, rows, start_row, start_col
        )

    def delete_rows_from_spreadsheet(
        self, spreadsheet_id, sheet_name, start_index, end_index
    ):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            spreadsheet_id
        )
        return self.infer_spreadsheet_storage(
            local_spreadsheet_id
        ).delete_rows_from_spreadsheet(
            local_spreadsheet_id, sheet_name, start_index, end_index
        )

    def delete_sheet(self, spreadsheet_id, sheet_name):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            spreadsheet_id
        )
        return self.infer_spreadsheet_storage(local_spreadsheet_id).delete_sheet(
            local_spreadsheet_id, sheet_name
        )

    def get_slide_elements(self, presentation_id):
        local_presentation_id = self.__add_local_storage_presentation_file_extension(
            presentation_id
        )
        return self.infer_slides_storage(local_presentation_id).get_slide_elements(
            local_presentation_id
        )

    def get_slides_metadata_re_string(self, presentation_id):
        local_spreadsheet_id = self.__add_local_storage_sheet_file_extension(
            presentation_id
        )
        return self.infer_slides_storage(
            local_spreadsheet_id
        ).get_slides_metadata_re_string()
